#!/usr/bin/env python3
#
# Copyright 2022 Max Planck Insitute Magdeburg
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
#
"""Functions for parsing and converting constraints and linear expressions"""

from typing import Dict, List, Tuple
from scipy import sparse
import re


def parse_constraints(constr, reaction_ids) -> list:
    """Parses linear constraints written as strings
    
    Parses one or more *linear* constraints written as strings.
    
    Args:
        constr (str or list of str): 
            (List of) constraints in string form.
            E.g.: ['r1 + 3*r2 = 0.3', '-5*r3 -r4 <= -0.5'] or
            '1.0 r1 + 3.0*r2 =0.3,-r4-5*r3<=-0.5' or ...
            
        reaction_ids (list of str): 
            List of reaction identifiers.

    Returns:
        (List of dicts): 
        List of constraints. Each constraint is a list of three elements.
        E.g.: [[{'r1':1.0,'r2':3.0},'=',0.3],[{'r3':-5.0,'r4':-1.0},'<=',-0.5],...]
    """
    if not constr:
        return []
    if type(constr) is str:
        if "\n" in constr or "," in constr:
            constr = re.split(r"\n|,", constr)
    if bool(constr) and (type(constr) is not list or type(constr[0]) is dict):
        constr = [constr]
    for i, c in enumerate(constr):
        if type(c) is tuple:
            constr[i] = list(c)
    if type(constr[0]) is not list:
        constr = lineq2list(constr, reaction_ids)
    return constr


def parse_linexpr(expr, reaction_ids) -> List:
    """Parses linear expressions written as strings
    
    Parses one or more *linear* expressions written as strings.
    
    Args:
        expr (str or list of str): 
            (List of) expressions in string form.
            E.g.: ['r1 + 3*r2', '-5*r3 -r4'] or
            '1.0 r1 + 3.0*r2,-r4-5*r3' or ...
            
        reaction_ids (list of str): 
            List of reaction identifiers.

    Returns:
        (List of dicts): 
        List of expressions. Each expression is a dictionary.
        E.g.: [{'r1':1.0,'r2':3.0},{'r3':-5.0,'r4':-1.0},...]
    """
    if not expr:
        return []
    if type(expr) is str:
        if "\n" in expr or "," in expr:
            expr = re.split(r"\n|,", expr)
    if bool(expr) and (type(expr) is not list):
        expr = [expr]
    return [linexpr2dict(e, reaction_ids) if type(e) is str else e for e in expr]


def lineq2mat(equations, reaction_ids) -> Tuple[sparse.csr_matrix, Tuple, sparse.csr_matrix, Tuple]:
    """Translates *linear* (in)equalities to matrices
    
    Input inequalities in the form of strings is translated into matrices and vectors. The reaction
    list defines the order of variables and thus the columns of the resulting matrices, the order
    of (in)equalities will be preserved in the output matrices. As an example, take the input:
    
    equations = ['2*c - b +3*a <= 2','c - b = 0','2*b -a >=-2'], reaction_ids = ['a','b','c']
    
    This will be translated to the form A_ineq * x <= b_ineq, A_eq * x = b_eq and hence to
    
    A_ineq = sparse.csr_matrix([[3,-1,2],[1,-2,0]]), b_ineq = [2,2],
    A_eq = sparse.csr_matrix([[1,-2,0]]), b_eq = [0]
    
    Args:
        equations (list of str): 
            (List of) (in)equalities in string form equations=['r1 + 3*r2 = 0.3', '-5*r3 -r4 <= -0.5']
            
        reaction_ids (list of str): 
            List of reaction identifiers or variable names that are used to recognize variables in
            the provided (in)equalities

    Returns:
        (Tuple): 
        A_ineq, b_ineq, A_eq, b_eq. Coefficient matrices and right hand sides that represent the input
        (in)equalities as matrix-vector multiplications
    """
    numr = len(reaction_ids)
    A_ineq = sparse.csr_matrix((0, numr))
    b_ineq = []
    A_eq = sparse.csr_matrix((0, numr))
    b_eq = []
    for equation in equations:
        try:
            lhs, rhs = re.split('<=|=|>=', equation)
            eq_sign = re.search('<=|>=|=', equation)[0]
            rhs = float(rhs)
        except:
            raise Exception("Equations must contain exactly one (in)equality sign: <=,=,>=. Right hand side must be a float number.")
        A = linexpr2mat(lhs, reaction_ids)
        if eq_sign == '=':
            A_eq = sparse.vstack((A_eq, A))
            b_eq += [rhs]
        elif eq_sign == '<=':
            A_ineq = sparse.vstack((A_ineq, A))
            b_ineq += [rhs]
        elif eq_sign == '>=':
            A_ineq = sparse.vstack((A_ineq, -A))
            b_ineq += [-rhs]
    return A_ineq, b_ineq, A_eq, b_eq


def lineq2list(equations, reaction_ids) -> List:
    """Translates *linear* (in)equalities to list format: [lhs,sign,rhs]
    
    Input inequalities in the form of strings are translated into a specific list format that
    facilitates the readout of left-hand-side, equality sign and right-hand-side of the inequality.
    
    equations = ['2*c - b +3*a <= 2','c - b = 0','2*b -5...], reaction_ids = ['a','b','c']
    
    This will be translated to the [[{'a':3.0,'b':-1.0,'c':2.0},'<=',2.0],[{'b':-1.0,'c':1.0},'=',0.0], ...]
    
    Args:
        equations (list of str): 
            (List of) (in)equalities in string form equations=['r1 + 3*r2 = 0.3', '-5*r3 -r4 <= -0.5']
            
        reaction_ids (list of str): 
            List of reaction identifiers or variable names that are used to recognize variables in
            the provided (in)equalities

    Returns:
        (list of lists): 
        (In)equalities presented in the form:
        [[{'a':3.0,'b':-1.0,'c':2.0},'<=',2.0], # e1
         [{'b':-1.0,'c':1.0},'=',0.0],          # e2
         ...]                                   # ...
    """
    D = []
    for equation in equations:
        if not equation:
            continue
        try:
            lhs, rhs = re.split('<=|=|>=', equation)
            eq_sign = re.search('<=|>=|=', equation)[0]
            rhs = float(rhs)
        except:
            raise Exception("Equations must contain exactly one (in)equality sign: <=,=,>=. Right hand side must be a float number.")
        D.append((linexpr2dict(lhs, reaction_ids), eq_sign, rhs))
    return D


def lineqlist2str(D):
    """Translates a *linear* (in)equality from the list format [lhs,sign,rhs] to a string
    
    E.g. input: D=[{'a':3.0,'b':-1.0,'c':2.0},'<=',2.0]] is translated to: out='3.0 a - 1.0 b + 2.0 c <= 2'
    
    Args:
        D (list): 
            (In)equality in list form, e.g.: D=[{'a':3.0,'b':-1.0,'c':2.0},'<=',2.0]]

    Returns:
        (str): 
        A list of (in)equalities in string form

    """
    if D[0]:
        return linexprdict2str(D[0]) + " " + D[1] + " " + str(D[2])
    elif D[1] and D[2]:
        return D[1] + " " + str(D[2])
    else:
        return ""


def lineqlist2mat(D, reaction_ids) -> Tuple[sparse.csr_matrix, Tuple, sparse.csr_matrix, Tuple]:
    """Translates *linear* (in)equalities presented in the list of lists format to matrices
    
    Input inequalities in the list of lists form is translated into matrices and vectors. The reaction
    list defines the order of variables and thus the columns of the resulting matrices, the order
    of (in)equalities will be preserved in the output matrices. As an example, take the input:
    
    D = [[{'a':3.0,'b':-1.0,'c':2.0},'<=',2.0],[{'b':-1.0,'c':1.0},'=',0.0], [{'a':-1,'b':2.0},'>=',-2.0]]
    
    This will be translated to the form A_ineq * x <= b_ineq, A_eq * x = b_eq and hence to
    
    A_ineq = sparse.csr_matrix([[3,-1,2],[1,-2,0]]), b_ineq = [2,2],
    A_eq = sparse.csr_matrix([[1,-2,0]]), b_eq = [0]
    
    Args:
        D (list of dict): 
            (List of) (in)equalities in the list of list form: 
            [[{'a':3.0,'b':-1.0,'c':2.0},'<=',2.0],[{'b':-1.0,'c':1.0},'=',0.0], ...]
            
        reaction_ids (list of str): 
            List of reaction identifiers or variable names that are used to recognize variables in
            the provided (in)equalities

    Returns:
        (Tuple): 
        A_ineq, b_ineq, A_eq, b_eq. Coefficient matrices and right hand sides that represent the input
        (in)equalities as matrix-vector multiplications
    """
    numr = len(reaction_ids)
    A_ineq = sparse.csr_matrix((0, numr))
    b_ineq = []
    A_eq = sparse.csr_matrix((0, numr))
    b_eq = []
    for d in D:
        d_expr = linexprdict2mat(d[0], reaction_ids)
        eq_sign = d[1]
        rhs = d[2]
        if eq_sign == '=':
            A_eq = sparse.vstack((A_eq, d_expr))
            b_eq += [rhs]
        elif eq_sign == '<=':
            A_ineq = sparse.vstack((A_ineq, d_expr))
            b_ineq += [rhs]
        elif eq_sign == '>=':
            A_ineq = sparse.vstack((A_ineq, -d_expr))
            b_ineq += [-rhs]
    return A_ineq, b_ineq, A_eq, b_eq


def linexpr2mat(expr, reaction_ids) -> sparse.csr_matrix:
    """Translates a linear expression into a vector
    
    E.g.: input: expr='2 R3 - R1', reaction_ids=['R1', 'R2', 'R3', 'R4'] translates into sparse matrix:  A = [-1 0 2 0]
    
    Args:
        expr (str): 
            (In)equality as a character string: e.g., expr='2 R3 - R1'
                        
        reaction_ids (list of str): 
            List of reaction identifiers or variable names that are used to recognize variables in the input

    Returns:
        (sparse.csr_matrix): 
        A single-row coefficient matrix that represents the input expression when multiplied with the variable vector
    """
    #
    A = sparse.lil_matrix((1, len(reaction_ids)))
    # split expression into parts and strip away special characters
    expr_parts = [re.sub(r'^(\s|-|\+|\()*|(\s|-|\+|\))*$', '', part) for part in expr.split()]
    # identify reaction identifiers by comparing with models reaction list
    ridx = [r for r in expr_parts if r in reaction_ids]
    # verify syntax of expression
    # 1. there must not be two numbers in a row
    # 2. there must not remain words that are not reaction identifiers
    # 3. there must be no reaction id duplicates
    last_was_number = False
    for part in expr_parts:
        if part in ridx:
            last_was_number = False
            continue
        if re.match('^\d*\.{0,1}\d*$', part) is not None:
            if last_was_number:
                raise Exception("Expression invalid. The expression contains at least two numbers in a row.")
            last_was_number = True
            continue
        raise Exception("Expression invalid. Unknown identifier " + part + ".")
    if not len(ridx) == len(set(ridx)):
        raise Exception("Reaction identifiers may only occur once in each linear expression.")
    # iterate through reaction identifiers and retrieve coefficients from linear expression
    for rid in ridx:
        coeff = re.search('(\s|^)(\s|\d|-|\+|\.)*?(?=' + re.escape(rid) + '(\s|$))', expr)[0]
        coeff = re.sub('\s', '', coeff)
        if coeff in ['', '+']:
            coeff = 1.0
        if coeff == '-':
            coeff = -1.0
        else:
            coeff = float(coeff)
        A[0, reaction_ids.index(rid)] = coeff
    return A.tocsr()


def linexpr2dict(expr, reaction_ids) -> dict:
    """Translates a linear expression into a dictionary
    
    E.g.: input: expr='2 R3 - R1', reaction_ids=['R1', 'R2', 'R3', 'R4'] translates to a dict D={'R1':-1.0, 'R3': 2.0}
    
    Args:
        expr (str): 
            (In)equalities as a character string, e.g.: expr='2 R3 - R1'
            
        reaction_ids (list of str): 
            List of reaction identifiers or variable names that are used to recognize variables in the input

    Returns:
        (dict): 
        A dictionary that contains the variable names and the variable coefficients in the linear expression
    """
    # split expression into parts and strip away special characters
    expr_parts = [re.sub(r'^(\s|-|\+|\()*|(\s|-|\+|\))*$', '', part) for part in expr.split()]
    expr_parts = [e for e in expr_parts if e != '']  # remove 'empty' entries
    # identify reaction identifiers by comparing with models reaction list
    ridx = [r for r in expr_parts if r in reaction_ids]
    # verify syntax of expression
    # 1. there must not be two numbers in a row
    # 2. there must not remain words that are not reaction identifiers
    # 3. there must be no reaction id duplicates
    last_was_number = False
    for part in expr_parts:
        if part in ridx:
            last_was_number = False
            continue
        if re.match('^\d*\.{0,1}\d*$', part) is not None:
            if last_was_number:
                raise Exception("Expression invalid. The expression contains at least two numbers in a row.")
            last_was_number = True
            continue
        raise Exception("Expression invalid. Unknown identifier " + part + ".")
    if not len(ridx) == len(set(ridx)):
        raise Exception("Reaction identifiers may only occur once in each linear expression.")
    D = {}
    # iterate through reaction identifiers and retrieve coefficients from linear expression
    for rid in ridx:
        coeff = re.search('(\s|^)(\s|\d|-|\+|\.)*?(?=' + re.escape(rid) + '(\s|$))', expr)[0]
        coeff = re.sub('\s', '', coeff)
        if coeff in ['', '+']:
            coeff = 1.0
        elif coeff == '-':
            coeff = -1.0
        else:
            coeff = float(coeff)
        D.update({rid: coeff})
    return D


def linexprdict2mat(D, reaction_ids) -> sparse.csr_matrix:
    """Translates a linear expression from dict into a matrix
    
    E.g.: input: D={'R1':-1.0, 'R3': 2.0}, reaction_ids=['R1', 'R2', 'R3', 'R4'] translates into sparse matrix:  A = [-1 0 2 0]
    
    Args:
        D (dict): 
            Linear expression as a dictionary
                        
        reaction_ids (list of str): 
            List of reaction identifiers or variable names

    Returns:
        (sparse.csr_matrix): 
        A single-row coefficient matrix that represents the input expression when multiplied with the variable vector
    """
    # iterate through reaction identifiers and retrieve coefficients from linear expression
    A = sparse.lil_matrix((1, len(reaction_ids)))
    for k, v in D.items():
        A[0, reaction_ids.index(k)] = v
    return A.tocsr()


def linexprdict2str(D):
    """Translates a linear expression from dict into a caracter string
    
    E.g.: input: D={'R1':-1.0, 'R3': 2.0}, translates to the string:  '- 1.0 R1 + 2.0 R3'
    
    Args:
        D (dict): 
            Linear expression as a dictionary

    Returns:
        (str): 
        The input linear expression as a character string
    """
    if D:
        expr_parts = [str(v) + " " + k for k, v in D.items()]
        expr = expr_parts[0]
        for ep in expr_parts[1:]:
            if ep[0] == '-':
                expr += ' - ' + ep[1:]
            else:
                expr += ' + ' + ep
        return expr
    else:
        return ""


def get_rids(expr, reaction_ids):
    """Get reaction identifiers that are present in string
    
    E.g.: input: D={'R1':-1.0, 'R3': 2.0}, translates to the string:  '- 1.0 R1 + 2.0 R3'
    
    Args:
        expr (str): 
            A character string
                        
        reaction_ids (list of str): 
            List of reaction identifiers or variable names

    Returns:
        (list of str): 
        A list of strings containing the reaction/variable strings present in the input string
    """
    expr_parts = [re.sub(r'^(\s|-|\+|\()*|(\s|-|\+|\|<|\=|>)*$', '', part) for part in expr.split()]
    reacIDs = []
    for part in expr_parts:
        if part in reaction_ids:
            reacIDs += [part]
            continue
        if re.match('^\d*\.{0,1}\d*$', part) is not None:
            continue
        raise Exception("Expression invalid. Unknown identifier " + part + ".")
    return reacIDs



#!/usr/bin/env python3
#
# Copyright 2022 Max Planck Insitute Magdeburg
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
#
#
"""Function: computing metabolic strain designs (compute_strain_designs)"""

from contextlib import redirect_stdout, redirect_stderr
from typing import Dict, List, Tuple
import numpy as np
import logging
import json
import io
from copy import deepcopy
import logging
from cobra import Model
from cobra.manipulation import rename_genes
from straindesign import SDModule, SDSolutions, select_solver, fva, DisableLogger, SDProblem, SDMILP
from straindesign.names import *
from straindesign.networktools import   remove_ext_mets, remove_dummy_bounds, bound_blocked_or_irrevers_fva, \
                                        remove_irrelevant_genes, extend_model_gpr, extend_model_regulatory, \
                                        compress_model, compress_modules, compress_ki_ko_cost, expand_sd, filter_sd_maxcost


def compute_strain_designs(model: Model, **kwargs: dict) -> SDSolutions:
    """Computes strain designs for a user-defined strain design problem

    A number of arguments can be specified to detail the problem and influence the solution process.
    This function supports the computation of Minimal Cut Sets (MCS), OptKock, RobustKnock and OptCouple
    strain designs. It is possible to combine any of the latter ones with the MCS approach, e.g., to
    engineer growth coupled production, but also suppress the production of an undesired by-product.
    The computation can be started in two different ways. Either by specifying the computation parameters
    indivdually or reuse a parameters dictionary from a previous computation. CNApy stores strain design
    setup dics as JSON ".sd"-files that can be loaded in python and used as an input for this function.

    Example:
        sols = compute_strain_designs(model, sd_modules=[sd_module1, sd_module2], solution_approach = 'any')

    Args:
        model (cobra.Model):
            A metabolic model that is an instance of the cobra.Model class. The model may or may not
            contain genes/GPR-rules.

        sd_setup (dict):
            sd_setup should be a dictionary containing a set of parameters for strain design computation.
            The allowed keywords are the same listed hereafter. Therefore, *sd_setup and other arguments
            (except for model) must not be used together*.

        sd_modules ([straindesign.SDModule]):
            List of strain design modules that describe the sub-problems, such as the MCS-like protection
            or suppression of flux subspaces or the OptKnock, RobustKnock or OptCouple objective and
            constraints. The list of modules determines the global objective function of the strain design
            computation. If only SUPPRESS and PROTECT modules are used, the strain design computation is
            MCS-like, such that the number of interventions is minimized. If a module for one of the nested
            optimization approaches is used, the global objective function is retrieved from this module.
            The number of SUPPRESS and PROTECT modules is unrestricted and can be combined with the other
            modules, however only one of the modules OPTKNOCK, ROBUSKNOCK and OPTCOUPLE may be used at a time.
            For details, see SDModule.

        solver (optional (str)): (Default: same as defined in model / COBRApy)
            The solver that should be used for preparing and carrying out the strain design computation.
            Allowed values are 'cplex', 'gurobi', 'scip' and 'glpk'.

        max_cost (optional (int)): (Default: inf):
            The maximum cost threshold for interventions. Every possible intervention is associated with a
            cost value (1, by default). Strain designs cannot exceed the max_cost threshold. Individual
            intervention cost factors may be defined through ki_cost, ko_cost, gki_cost, gko_cost and reg_cost.

        max_solutions (optional (int)): (Default: inf)
            The maximum number of MILP solutions that are generated for a strain design problem. The number of returned
            strain designs is usually larger than the number of max_solutions, since a MILP solution is decompressed
            to multiple strain designs. When the compress-flag is set to 'False' the number of returned solutions is
            equal to max_solutions.

        M (optional (int)): (Default: None)
            If this value is specified (and non-zero, not None), the computation uses the big-M
            method instead of indicator constraints. Since GLPK does not support indicator constraints it uses
            the big-M method by default (with M=1000). M should be chosen 'sufficiently large' to avoid computational
            artifacts and 'sufficiently small' to avoid numerical issues.

        compress (optional (bool)): (Default: True)
            If 'True', the interative network compressor is used.

        gene_kos (optional (bool)): (Default: False)
            If 'True', strain designs are computed based on gene-knockouts instead of reaction knockouts. This
            parameter needs not be defined if any of ki_cost, ko_cost, gki_cost, gko_cost and reg_cost is used.
            By default, reactions are considered as knockout targets.

        ko_cost (optional (dict)): (Default: None)
            A dictionary of reaction identifiers and their associated knockout costs. If not specified, all reactions
            are treated as knockout candidates, equivalent to ko_cost = {'r1':1, 'r2':1, ...}. If a subset of reactions
            is listed in the dict, all other are not considered as knockout candidates.

        ki_cost (optional (dict)): (Default: None)
            A dictionary of reaction identifiers and their associated costs for addition. If not specified, all reactions
            are treated as knockout candidates. Reaction addition candidates must be present in the original model with
            the intended flux boundaries **after** insertion. Additions are treated adversely to knockouts, meaning that
            their exclusion from the network is not associated with any cost while their presence entails intervention costs.

        gko_cost (optional (dict)): (Default: None)
            A dictionary of gene identifiers and their associated knockout costs. To reference genes, gene IDs can be
            used,as well as gene names. If not specified, genes are not treated as knockout candidates. An exception is
            the 'gene_kos' argument. If 'gene_kos' is used, all genes are treated as knockout candidates with intervention
            costs of 1. This is equivalent to gko_cost = {'g1':1, 'g2':1, ...}.

        gki_cost (optional (dict)): (Default: None)
            A dictionary of gene identifiers and their associated addition costs. To reference genes, gene IDs can be
            used, as well as gene names. If not specified, none of the genes are treated as addition candidates.

        reg_cost (optional [dict]): ( Default: None)
            Regulatory interventions candidates can be optionally specified as a list. Thereby, the constraint marking the
            regulatory intervention is put as key and the associated intervention cost is used as the corresponding value.
            E.g., reg_cost = {'1 EX_o2_e = -1': 1, ... <other regulatory interventions>}. Instead of strings, constraints
            can also be passed as lists. reg_cost = {[{'EX_o2_e':1}, '=', -1]: 1, ...}

        solution_approach (optional (str)): ( Default: 'best')
            The approach used to find strain designs. Possible values are 'any', 'best' or 'populate'. 'any' is usually the
            fastest option, since optimality is not enforced. Hereby computed MCS are still irreducible intervention sets,
            however, not MCS with the fewest possible number of interventions. 'best' computes globally optimal strain designs,
            that is, MCS with the fewest number of interventions, OptKnock strain designs with the highest possible production
            rate, OptCouple strain designs with the hightest growth coupling potential etc.. 'populate' does the same as 'best',
            but makes use of CPLEX' and Gurobi's populate function to generate multiple strain designs. It is identical to 'best'
            when used with SCIP or GLPK.
            Attention:
            If 'any' used with OptKnock, for instance, the MILP may return the wild type as a possible immediately. Technically,
            the wiltype fulfills the criterion of maximal growth (inner objective) and maximality of the global objective is
            omitted by using 'any', so that carrying no product synthesis is permitted. Additional constraints can be used
            in the OptKnock problem to circumvent this. However, Optknock should generally be used with the 'best' option.

        time_limit (optional (int)): (Default: inf)
            The time limit in seconds for the MILP-solver.

        advanced, use_scenario (optional (bool)):
            Dummy parameters used for the CNApy interface.

    Returns:
        (SDSolutions):

            An object that contains all computed strain designs. If strain designs were computed
            as gene-interventions, the solution object contains a set of corresponding reaction-interventions
            that facilitate the analysis of the computed strain designs with COBRA methods.
    """
    allowed_keys = {
        MODULES, SETUP, SOLVER, MAX_COST, MAX_SOLUTIONS, 'M', 'compress', 'gene_kos', KOCOST, KICOST, GKOCOST, GKICOST, REGCOST,
        SOLUTION_APPROACH, 'advanced', 'use_scenario', T_LIMIT, SEED
    }
    logging.info('Preparing strain design computation.')
    if SETUP in kwargs:
        if type(kwargs[SETUP]) is str:
            with open(kwargs[SETUP], 'r') as fs:
                kwargs = json.load(fs)
        else:
            kwargs = kwargs[SETUP]

    if MODULES in kwargs:
        sd_modules = kwargs.pop(MODULES)
        if isinstance(sd_modules, SDModule):
            sd_modules = [sd_modules]
        orig_sd_modules = [m.copy() for m in sd_modules]
        sd_modules = [m.copy() for m in sd_modules]

    if SOLVER in kwargs:
        kwargs[SOLVER] = select_solver(kwargs[SOLVER])
    else:
        kwargs[SOLVER] = select_solver(None, model)

    if MAX_COST in kwargs:
        kwargs.update({MAX_COST: float(kwargs.pop(MAX_COST))})
    else:
        kwargs.update({MAX_COST: np.inf})

    if MODEL_ID in kwargs:
        model_id = kwargs.pop(MODEL_ID)
        if model_id != model.id:
            logging.warning( "Model IDs of provided model and setup not matching. Apparently, ",\
                          "the strain design setup was specified for a different model. "+\
                          "Errors might occur due to non-matching reaction or gene-identifiers.")

    if 'advanced' in kwargs:
        model_id = kwargs.pop('advanced')
    if 'use_scenario' in kwargs:
        model_id = kwargs.pop('use_scenario')

    if SEED not in kwargs:
        kwargs[SEED] = int(np.random.default_rng().integers(1, 2**16 - 1))
        logging.info("  Using random seed " + str(kwargs[SEED]))
    else:
        logging.info("  Using seed " + str(kwargs[SEED]))

    # check all keys passed in kwargs
    for key, value in dict(kwargs).items():
        if key not in allowed_keys:
            raise Exception("Key " + key + " is not supported.")
        if key == KOCOST:
            uncmp_ko_cost = value
        if key == KICOST:
            uncmp_ki_cost = value
        if key == GKOCOST:
            uncmp_gko_cost = value
        if key == GKICOST:
            uncmp_gki_cost = value
        if key == REGCOST:
            uncmp_reg_cost = value
    if (GKOCOST in kwargs or GKICOST in kwargs or
        ('gene_kos' in kwargs and kwargs['gene_kos'])) and hasattr(model, 'genes') and model.genes:
        kwargs['gene_kos'] = True
        used_g_ids = set(kwargs[GKOCOST] if GKOCOST in kwargs and kwargs[GKOCOST] else set())
        used_g_ids.update(set(kwargs[GKICOST] if GKICOST in kwargs and kwargs[GKICOST] else set()))
        used_g_ids.update(set(kwargs[REGCOST] if REGCOST in kwargs and kwargs[REGCOST] else set()))
        # genes must not begin with number, put a 'g' in front of genes that start with a number
        if any([True for g in model.genes if g.id[0].isdigit()]):
            logging.warning("Gene IDs must not start with a digit. Inserting prefix 'g' where necessary.")
            rename_genes(model, {g.id: 'g' + g.id for g in model.genes if g.id[0].isdigit()})
        if np.all([len(g.name) for g in model.genes]) and (np.any([g.name in used_g_ids for g in model.genes]) or not used_g_ids):
            has_gene_names = True
        else:
            has_gene_names = False
        if has_gene_names and any([True for g in model.genes if g.name[0].isdigit()]):
            logging.warning("Gene names must not start with a digit. Inserting prefix 'g' where necessary.")
            for g, v in {g.id: 'g' + g.name for g in model.genes if g.name[0].isdigit()}.items():
                model.genes.get_by_id(g).name = v
        if GKOCOST not in kwargs or not kwargs[GKOCOST]:
            if has_gene_names:  # if gene names are defined, use them instead of ids
                uncmp_gko_cost = {k: 1.0 for k in model.genes.list_attr('name')}
            else:
                uncmp_gko_cost = {k: 1.0 for k in model.genes.list_attr('id')}
        if GKICOST not in kwargs or not kwargs[GKICOST]:
            uncmp_gki_cost = {}
    else:
        kwargs['gene_kos'] = False
        has_gene_names = False
    if KOCOST not in kwargs and not kwargs['gene_kos']:
        uncmp_ko_cost = {k: 1.0 for k in model.reactions.list_attr('id')}
    elif KOCOST not in kwargs or not kwargs[KOCOST]:
        uncmp_ko_cost = {}
    if KICOST not in kwargs or not kwargs[KICOST]:
        uncmp_ki_cost = {}
    if REGCOST not in kwargs or not kwargs[REGCOST]:
        uncmp_reg_cost = {}
    # check that at most one bilevel module is provided
    bilvl_modules = [i for i,m in enumerate(sd_modules) \
                if m[MODULE_TYPE] in [OPTKNOCK,ROBUSTKNOCK,OPTCOUPLE]]
    if len(bilvl_modules) > 1:
        raise Exception("Only one of the module types 'OptKnock', 'RobustKnock' and 'OptCouple' can be defined per "\
                            "strain design setup.")
    logging.info('  Using ' + kwargs[SOLVER] + ' for solving LPs during preprocessing.')
    with redirect_stdout(io.StringIO()), redirect_stderr(io.StringIO()), DisableLogger():  # suppress standard output from copying model
        orig_model = model
        model = model.copy()
        uncmp_model = model.copy()
    orig_ko_cost = deepcopy(uncmp_ko_cost)
    orig_ki_cost = deepcopy(uncmp_ki_cost)
    orig_reg_cost = deepcopy(uncmp_reg_cost)
    if 'compress' not in kwargs:
        kwargs['compress'] = True
    if kwargs['gene_kos']:
        orig_gko_cost = uncmp_gko_cost
        orig_gki_cost = uncmp_gki_cost
        # ensure that gene and reaction kos/kis do not overlap
        g_itv = {g for g in list(uncmp_gko_cost.keys()) + list(uncmp_gki_cost.keys())}
        r_itv = {r for r in list(uncmp_ko_cost.keys()) + list(uncmp_ki_cost.keys())}
        if np.any([np.any([True for g in uncmp_model.reactions.get_by_id(r).genes if g in g_itv]) for r in r_itv]) or \
            np.any(set(uncmp_gko_cost.keys()).intersection(set(uncmp_gki_cost.keys()))) or \
            np.any(set(uncmp_ko_cost.keys()).intersection(set(uncmp_ki_cost.keys()))):
            raise Exception('Specified gene and reaction knock-out/-in costs contain overlap. '\
                            'Make sure that metabolic interventions are enabled either through reaction or '\
                            'through gene interventions and are defined either as knock-ins or as knock-outs.')
    # 1) Preprocess Model
    with redirect_stdout(io.StringIO()), redirect_stderr(io.StringIO()), DisableLogger():  # suppress standard output from copying model
        cmp_model = uncmp_model.copy()
    # remove external metabolites
    remove_ext_mets(cmp_model)
    # replace model bounds with +/- inf if above a certain threshold
    remove_dummy_bounds(model)
    # FVAs to identify blocked, irreversible and essential reactions, as well as non-bounding bounds
    logging.info('  FVA to identify blocked reactions and irreversibilities.')
    bound_blocked_or_irrevers_fva(model, solver=kwargs[SOLVER])
    logging.info('  FVA(s) to identify essential reactions.')
    essential_reacs = set()
    for m in sd_modules:
        if m[MODULE_TYPE] != SUPPRESS:  # Essential reactions can only be determined from desired
            # or opt-/robustknock modules
            flux_limits = fva(cmp_model, solver=kwargs[SOLVER], constraints=m[CONSTRAINTS])
            for (reac_id, limits) in flux_limits.iterrows():
                if np.min(abs(limits)) > 1e-10 and np.prod(np.sign(limits)) > 0:  # find essential
                    essential_reacs.add(reac_id)
    # remove ko-costs (and thus knockability) of essential reactions
    [uncmp_ko_cost.pop(er) for er in essential_reacs if er in uncmp_ko_cost]
    # If computation of gene-gased intervention strategies, (optionally) compress gpr are rules and extend stoichimetric network with genes
    if kwargs['gene_kos']:
        if kwargs['compress'] is True or kwargs['compress'] is None:
            num_genes = len(cmp_model.genes)
            num_gpr = len([True for r in model.reactions if r.gene_reaction_rule])
            logging.info('Preprocessing GPR rules (' + str(num_genes) + ' genes, ' + str(num_gpr) + ' gpr rules).')
            # removing irrelevant genes will also remove essential reactions from the list of knockable genes
            uncmp_gko_cost = remove_irrelevant_genes(cmp_model, essential_reacs, uncmp_gki_cost, uncmp_gko_cost)
            if len(cmp_model.genes) < num_genes or len([True for r in model.reactions if r.gene_reaction_rule]) < num_gpr:
                num_genes = len(cmp_model.genes)
                num_gpr = len([True for r in cmp_model.reactions if r.gene_reaction_rule])
                logging.info('  Simplifyied to '+str(num_genes)+' genes and '+\
                    str(num_gpr)+' gpr rules.')
        logging.info('  Extending metabolic network with gpr associations.')
        reac_map = extend_model_gpr(cmp_model, has_gene_names)
        for i, m in enumerate(sd_modules):
            for p in [CONSTRAINTS, INNER_OBJECTIVE, OUTER_OBJECTIVE, PROD_ID]:
                if p in m and m[p] is not None:
                    if p == CONSTRAINTS:
                        for c in m[p]:
                            for k in list(c[0].keys()):
                                v = c[0].pop(k)
                                for n, w in reac_map[k].items():
                                    c[0][n] = v * w
                    if p in [INNER_OBJECTIVE, OUTER_OBJECTIVE, PROD_ID]:
                        for k in list(m[p].keys()):
                            v = m[p].pop(k)
                            for n, w in reac_map[k].items():
                                m[p][n] = v * w
        uncmp_ko_cost.update(uncmp_gko_cost)
        uncmp_ki_cost.update(uncmp_gki_cost)
    uncmp_ko_cost.update(extend_model_regulatory(cmp_model, uncmp_reg_cost))
    cmp_ko_cost = uncmp_ko_cost
    cmp_ki_cost = uncmp_ki_cost
    # Compress model
    if kwargs['compress'] is True or kwargs['compress'] is None:  # If compression is activated (or not defined)
        logging.info('Compressing Network (' + str(len(cmp_model.reactions)) + ' reactions).')
        # compress network by lumping sequential and parallel reactions alternatingly.
        # Exclude reactions named in strain design modules from parallel compression
        no_par_compress_reacs = set()
        for m in sd_modules:
            for p in [CONSTRAINTS, INNER_OBJECTIVE, OUTER_OBJECTIVE, PROD_ID]:
                if p in m and m[p] is not None:
                    param = m[p]
                    if p == CONSTRAINTS:
                        for c in param:
                            for k in c[0].keys():
                                no_par_compress_reacs.add(k)
                    if p in [INNER_OBJECTIVE, OUTER_OBJECTIVE, PROD_ID]:
                        for k in param.keys():
                            no_par_compress_reacs.add(k)
        cmp_mapReac = compress_model(cmp_model, no_par_compress_reacs)
        # compress information in strain design modules
        sd_modules = compress_modules(sd_modules, cmp_mapReac)
        # compress ko_cost and ki_cost
        cmp_ko_cost, cmp_ki_cost, cmp_mapReac = compress_ki_ko_cost(cmp_ko_cost, cmp_ki_cost, cmp_mapReac)
    else:
        cmp_mapReac = []

    # An FVA to identify essentials before building and launching MILP (not sure if this has an effect)
    logging.info('  FVA(s) in compressed model to identify essential reactions.')
    essential_reacs = set()
    for m in sd_modules:
        if m[MODULE_TYPE] != SUPPRESS:  # Essential reactions can only be determined from desired
            # or opt-/robustknock modules
            flux_limits = fva(cmp_model, solver=kwargs[SOLVER], constraints=m[CONSTRAINTS])
            for (reac_id, limits) in flux_limits.iterrows():
                if np.min(abs(limits)) > 1e-10 and np.prod(np.sign(limits)) > 0:  # find essential
                    essential_reacs.add(reac_id)
    # remove ko-costs (and thus knockability) of essential reactions
    [cmp_ko_cost.pop(er) for er in essential_reacs if er in cmp_ko_cost]
    essential_kis = set(cmp_ki_cost[er] for er in essential_reacs if er in cmp_ki_cost)
    # Build MILP
    kwargs1 = kwargs
    kwargs1[KOCOST] = cmp_ko_cost
    kwargs1[KICOST] = cmp_ki_cost
    kwargs1['essential_kis'] = essential_kis
    kwargs1.pop('compress')
    if GKOCOST in kwargs1:
        kwargs1.pop(GKOCOST)
    if GKICOST in kwargs1:
        kwargs1.pop(GKICOST)
    if REGCOST in kwargs1:
        kwargs1.pop(REGCOST)

    kwargs_milp = {k: v for k, v in kwargs.items() if k in [SOLVER, MAX_COST, 'M', SEED]}
    kwargs_milp.update({KOCOST: cmp_ko_cost})
    kwargs_milp.update({KICOST: cmp_ki_cost})
    kwargs_milp.update({'essential_kis': essential_kis})
    logging.info("Finished preprocessing:")
    logging.info("  Model size: " + str(len(cmp_model.reactions)) + " reactions, " + str(len(cmp_model.metabolites)) + " metabolites")
    logging.info("  " + str(len(cmp_ko_cost) + len(cmp_ki_cost) - len(essential_kis)) + " targetable reactions")

    sd_milp = SDMILP(cmp_model, sd_modules, **kwargs_milp)

    kwargs_computation = {}
    if MAX_SOLUTIONS in kwargs:
        kwargs_computation.update({MAX_SOLUTIONS: float(kwargs.pop(MAX_SOLUTIONS))})
    if T_LIMIT in kwargs:
        kwargs_computation.update({T_LIMIT: float(kwargs.pop(T_LIMIT))})
    kwargs_computation.update({'show_no_ki': True})

    # solution approach
    if SOLUTION_APPROACH in kwargs:
        solution_approach = kwargs.pop(SOLUTION_APPROACH)
    else:
        solution_approach = BEST

    # solve MILP
    if solution_approach == ANY:
        cmp_sd_solution = sd_milp.compute(**kwargs_computation)
    elif solution_approach == BEST:
        cmp_sd_solution = sd_milp.compute_optimal(**kwargs_computation)
    elif solution_approach == POPULATE:
        cmp_sd_solution = sd_milp.enumerate(**kwargs_computation)

    logging.info('  Decompressing.')
    if cmp_sd_solution.status in [OPTIMAL, TIME_LIMIT_W_SOL]:
        sd = expand_sd(cmp_sd_solution.get_reaction_sd_mark_no_ki(), cmp_mapReac)
        sd = filter_sd_maxcost(sd, kwargs[MAX_COST], uncmp_ko_cost, uncmp_ki_cost)
        sd = postprocess_reg_sd(uncmp_reg_cost, sd)
    else:
        sd = []

    setup = deepcopy(cmp_sd_solution.sd_setup)
    setup.update({MODULES: orig_sd_modules, KOCOST: orig_ko_cost, KICOST: orig_ki_cost, REGCOST: orig_reg_cost})
    if kwargs['gene_kos']:
        setup.update({GKOCOST: orig_gko_cost, GKICOST: orig_gki_cost})
    sd_solutions = SDSolutions(orig_model, sd, cmp_sd_solution.status, setup)
    logging.info(str(len(sd)) + ' solutions found.')

    return sd_solutions


def postprocess_reg_sd(reg_cost, sd):
    """Postprocess regulatory interventions

    Mark regulatory interventions with true or false"""
    for s in sd:
        for k, v in reg_cost.items():
            if k in s:
                s.pop(k)
                s.update({v['str']: True})
            else:
                s.update({v['str']: False})
    return sd



#!/usr/bin/env python3
#
# Copyright 2022 Max Planck Insitute Magdeburg
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
#
"""Gurobi solver interface for LP and MILP"""

from scipy import sparse
from numpy import nan, inf, isinf, sum, array, random
import gurobipy as gp
from gurobipy import GRB as grb
from straindesign.names import *
from typing import Tuple, List
import logging

gstatus = gp.StatusConstClass


class Gurobi_MILP_LP(gp.Model):
    """Gurobi interface for MILP and LP
    
    This class is a wrapper for the Gurobi-Python API to offer bindings and namings
    for functions for the construction and manipulation of MILPs and LPs in an
    vector-matrix-based manner that are consistent with those of the other solver 
    interfaces in the StrainDesign package. The purpose is to unify the instructions 
    for operating with MILPs and LPs throughout StrainDesign.
    
    The Gurobi interface provides support for indicator constraints as well as for
    the populate function.
    
    Accepts a (mixed integer) linear problem in the form:
        minimize(c),
        subject to: 
        A_ineq * x <= b_ineq,
        A_eq * x  = b_eq,
        lb <= x <= ub,
        forall(i) type(x_i) = vtype(i) (continous, binary, integer),
        indicator constraints:
        x(j) = [0|1] -> a_indic * x [<=|=|>=] b_indic
                
    Please ensure that the number of variables and (in)equalities is consistent
        
    Example: 
        gurobi = Gurobi_MILP_LP(c, A_ineq, b_ineq, A_eq, b_eq, lb, ub, vtype, indic_constr)
                
    Args:
        c (list of float): (Default: None)
            The objective vector (Objective sense: minimization).
            
        A_ineq (sparse.csr_matrix): (Default: None)
            A coefficient matrix of the static inequalities.   
            
        b_ineq (list of float): (Default: None)
            The right hand side of the static inequalities.
            
        A_eq (sparse.csr_matrix): (Default: None)
            A coefficient matrix of the static equalities.   
            
        b_eq (list of float): (Default: None)
            The right hand side of the static equalities.
            
        lb (list of float): (Default: None)
            The lower variable bounds.
            
        ub (list of float): (Default: None)
            The upper variable bounds.
            
        vtype (str): (Default: None)
            A character string that specifies the type of each variable:
            'c'ontinous, 'b'inary or 'i'nteger
            
        indic_constr (IndicatorConstraints): (Default: None)
            A set of indicator constraints stored in an object of IndicatorConstraints
            (see reference manual or docstring).
            
        seed (int16): (Default: None)
            An integer value serving as a seed to make MILP solving reproducible.
            
        Returns:
            (Gurobi_MILP_LP):
            
            A Gurobi MILP/LP interface class.
    """

    def __init__(self, c=None, A_ineq=None, b_ineq=None, A_eq=None, b_eq=None, lb=None, ub=None, vtype=None, indic_constr=None, seed=None):
        super().__init__()
        try:
            numvars = A_ineq.shape[1]
        except:
            numvars = A_eq.shape[1]
        # prepare coefficient matrix
        if isinstance(A_eq, list):
            if not A_eq:
                A_eq = sparse.csr_matrix((0, numvars))
        if isinstance(A_ineq, list):
            if not A_ineq:
                A_ineq = sparse.csr_matrix((0, numvars))

        for i, v in enumerate(b_ineq):
            if isinf(v):
                b_ineq[i] = grb.INFINITY
        # concatenate right hand sides
        # construct Gurobi problem. Add variables and linear constraints
        x = self.addMVar(len(c), obj=c, lb=lb, ub=ub, vtype=[k for k in vtype])
        self.setObjective(array(c) @ x, grb.MINIMIZE)
        if A_ineq.shape[0]:
            self.addMConstr(A_ineq, x, grb.LESS_EQUAL, array(b_ineq))
        if A_eq.shape[0]:
            self.addMConstr(A_eq, x, grb.EQUAL, array(b_eq))

        # add indicator constraints
        if not indic_constr == None:
            for i in range(len(indic_constr.sense)):
                self.addGenConstrIndicator(x[indic_constr.binv[i]],
                                        bool(indic_constr.indicval[i]),
                                        sum([indic_constr.A[i,j] * x[j] \
                                            for j in range(len(c)) if not indic_constr.A[i,j] == 0.0]),
                                        '=' if indic_constr.sense[i] =='E' else '<',
                                        indic_constr.b[i])

        # set parameters
        self.params.OutputFlag = 0
        self.params.OptimalityTol = 1e-9
        self.params.FeasibilityTol = 1e-9
        if 'B' in vtype or 'I' in vtype:
            if seed is None:
                # seed = random(0, grb.MAXINT)
                seed = int(random.randint(0, 2**16 - 1))
                logging.info('  MILP Seed: ' + str(seed))
            self.params.Seed = seed
            self.params.IntFeasTol = 1e-9  # (0 is not allowed by Gurobi)
            # yield only optimal solutions in pool
            self.params.PoolGap = 1e-9
            self.params.PoolGapAbs = 1e-9
            self.params.MIPFocus = 0
        self.update()

    def solve(self) -> Tuple[List, float, float]:
        """Solve the MILP or LP
        
        Example:
            sol_x, optim, status = gurobi.solve()
        
        Returns:
            (Tuple[List, float, float])
            
            solution_vector, optimal_value, optimization_status
        """
        try:
            self.optimize()  # call parent solve function (that was overwritten in this class)
            status = self.Status
            if status in [gstatus.OPTIMAL, gstatus.SOLUTION_LIMIT, gstatus.SUBOPTIMAL, gstatus.USER_OBJ_LIMIT]:  # solution
                min_cx = self.ObjVal
                status = OPTIMAL
            elif status == gstatus.TIME_LIMIT and not hasattr(self._Model__vars[0], 'X'):  # timeout without solution
                x = [nan] * self.NumVars
                min_cx = nan
                status = TIME_LIMIT
                return x, min_cx, status
            elif status == gstatus.TIME_LIMIT and hasattr(self._Model__vars[0], 'X'):
                min_cx = self.ObjVal
                status = TIME_LIMIT_W_SOL
            elif status in [gstatus.INF_OR_UNBD, gstatus.UNBOUNDED, gstatus.INFEASIBLE]:
                # solve problem again without objective to verify that problem is feasible
                self.params.DualReductions = 0
                self.optimize()
                self.params.DualReductions = 1
                if self.Status == gstatus.INFEASIBLE:
                    x = [nan] * self.NumVars
                    min_cx = nan
                    status = INFEASIBLE
                    return x, min_cx, status
                else:
                    x = [nan] * self.NumVars
                    min_cx = -inf
                    status = UNBOUNDED
                    return x, min_cx, status
            else:
                raise Exception('Status code ' + str(status) + " not yet handeld.")
            x = self.getSolution()
            return x, min_cx, status

        except gp.GurobiError as e:
            logging.error('Error code ' + str(e.errno) + ": " + str(e))
            min_cx = nan
            x = [nan] * self.NumVars
            return x, min_cx, ERROR

    def slim_solve(self) -> float:
        """Solve the MILP or LP, but return only the optimal value
                
        Example:
            optim = gurobi.slim_solve()
        
        Returns:
            (float)
            
            Optimum value of the objective function.
        """
        try:
            self.optimize()  # call parent solve function (that was overwritten in this class)
            status = self.Status
            if status in [gstatus.OPTIMAL, gstatus.SOLUTION_LIMIT, gstatus.SUBOPTIMAL,
                          gstatus.USER_OBJ_LIMIT]:  # solution integer optimal (tolerance)
                opt = self.ObjVal
            elif status in [gstatus.INF_OR_UNBD, gstatus.UNBOUNDED]:
                opt = -inf
            elif status in [gstatus.INFEASIBLE, gstatus.TIME_LIMIT]:
                opt = nan
            else:
                raise Exception('Status code ' + str(status) + " not yet handeld.")
            return opt
        except gp.GurobiError as e:
            logging.error('Error code ' + str(e.errno) + ": " + str(e))
            return nan

    def populate(self, n) -> Tuple[List, float, float]:
        """Generate a solution pool for MILPs
                
        Example:
            sols_x, optim, status = cplex.populate()
        
        Returns:
            (Tuple[List of lists, float, float])
            
            solution_vectors, optimal_value, optimization_status
        """
        try:
            if isinf(n):
                self.params.PoolSolutions = grb.MAXINT
            else:
                self.params.PoolSolutions = n
            self.params.PoolSearchMode = 2
            self.params.NumericFocus = 2
            self.optimize()  # call parent solve function (that was overwritten in this class)
            self.params.PoolSearchMode = 0
            self.params.NumericFocus = 0
            status = self.Status
            if status in [2, 10, 13, 15]:  # solution integer optimal
                min_cx = self.ObjVal
                status = OPTIMAL
            elif status == 9 and not hasattr(self._Model__vars[0], 'X'):  # timeout without solution
                x = [nan] * len(self._Model__vars)
                min_cx = nan
                status = TIME_LIMIT
                return x, min_cx, status
            elif status == 3:  # infeasible
                x = [nan] * len(self._Model__vars)
                min_cx = nan
                status = INFEASIBLE
                return x, min_cx, status
            elif status == 9 and hasattr(self._Model__vars[0], 'X'):  # timeout with solution
                min_cx = self.ObjVal
                status = TIME_LIMIT_W_SOL
            elif status in [4, 5]:  # solution unbounded
                min_cx = -inf
                x = [nan] * len(self._Model__vars)
                status = UNBOUNDED
                return x, min_cx, status
            else:
                raise Exception('Status code ' + str(status) + " not yet handeld.")
            x = self.getSolutions()
            return x, min_cx, status

        except gp.GurobiError as e:
            self.params.PoolSearchMode = 0
            logging.error('Error code ' + str(e.errno) + ": " + str(e))
            min_cx = nan
            x = [nan] * len(self._Model__vars)
            return x, min_cx, ERROR

    def set_objective(self, c):
        """Set the objective function with a vector"""
        for i in range(len(self._Model__vars)):
            self._Model__vars[i].Obj = c[i]
        self.update()
        if any([self._Model__vars[i].Obj for i in range(len(self._Model__vars))]):
            self.params.MIPFocus = 0
        else:
            self.params.MIPFocus = 1

    def set_objective_idx(self, C):
        """Set the objective function with index-value pairs
        
        e.g.: C=[[1, 1.0], [4,-0.2]]"""
        for c in C:
            self._Model__vars[c[0]].Obj = c[1]
        self.update()
        if any([self._Model__vars[i].Obj for i in range(len(self._Model__vars))]):
            self.params.MIPFocus = 0
        else:
            self.params.MIPFocus = 1

    def set_ub(self, ub):
        """Set the upper bounds to a given vector"""
        for i in range(len(ub)):
            self._Model__vars[ub[i][0]].ub = ub[i][1]
        self.update()

    def set_time_limit(self, t):
        """Set the computation time limit (in seconds)"""
        self.params.TimeLimit = t
        self.update()

    def add_ineq_constraints(self, A_ineq, b_ineq):
        """Add inequality constraints to the model
        
        Additional inequality constraints have the form A_ineq * x <= b_ineq.
        The number of columns in A_ineq must match with the number of variables x
        in the problem.
        
        Args:
            A_ineq (sparse.csr_matrix):
                The coefficient matrix
                
            b_ineq (list of float):
                The right hand side vector
        """
        vars = self._Model__vars
        for i in range(A_ineq.shape[0]):
            self.addLConstr(sum([A_ineq[i, j] * vars[j] for j in range(len(vars)) if not A_ineq[i, j] == 0.0]), grb.LESS_EQUAL, b_ineq[i])
        self.update()

    def add_eq_constraints(self, A_eq, b_eq):
        """Add equality constraints to the model
        
        Additional equality constraints have the form A_eq * x = b_eq.
        The number of columns in A_eq must match with the number of variables x
        in the problem.
        
        Args:
            A_eq (sparse.csr_matrix):
                The coefficient matrix
                
            b_eq (list of float):
                The right hand side vector
        """
        vars = self._Model__vars
        for i in range(A_eq.shape[0]):
            self.addLConstr(sum([A_eq[i, j] * vars[j] for j in range(len(vars)) if not A_eq[i, j] == 0.0]), grb.EQUAL, b_eq[i])
        self.update()

    def set_ineq_constraint(self, idx, a_ineq, b_ineq):
        """Replace a specific inequality constraint
        
        Replace the constraint with the index idx with the constraint a_ineq*x ~ b_ineq
        
        Args:
            idx (int):
                Index of the constraint
                
            a_ineq (list of float):
                The coefficient vector
                
            b_ineq (float):
                The right hand side value
        """
        constr = self._Model__constrs[idx]
        [self.chgCoeff(constr, x, val) for x, val in zip(self._Model__vars, a_ineq)]
        if isinf(b_ineq):
            constr.rhs = grb.INFINITY
        else:
            constr.rhs = b_ineq
        self.update()

    def getSolution(self) -> list:
        """Retrieve solution from Gurobi backend"""
        return [x.X for x in self._Model__vars]

    def getSolutions(self) -> list:
        """Retrieve solution pool from Gurobi backend"""
        nSols = self.SolCount
        x = []
        for i in range(nSols):
            self.setParam(grb.Param.SolutionNumber, i)
            if self.PoolObjVal == self.ObjVal:
                x += [[x.Xn for x in self._Model__vars]]
        return x



#!/usr/bin/env python3
#
# Copyright 2022 Max Planck Insitute Magdeburg
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
#
"""StrainDesign package for computational metabolic engineering"""

from importlib.util import find_spec as module_exists
from .names import *
import logging


class DisableLogger():
    """Environment in which logging is disabled"""

    def __enter__(self):
        logging.disable(logging.CRITICAL)

    def __exit__(self, exit_type, exit_value, exit_traceback):
        logging.disable(logging.NOTSET)


avail_solvers = set()
if module_exists("swiglpk"):
    avail_solvers.add(GLPK)
if module_exists("cplex"):
    avail_solvers.add(CPLEX)
if module_exists("gurobipy"):
    avail_solvers.add(GUROBI)
if module_exists("pyscipopt"):
    avail_solvers.add(SCIP)

from .solver_interface import *
from .indicatorConstraints import *
from .pool import *
from .efmtool import *
from .parse_constr import *
from .lptools import *
from .networktools import *
from .strainDesignModule import *
from .strainDesignSolutions import *
from .strainDesignProblem import *
from .strainDesignMILP import *
from .compute_strain_designs import *



"""Test if basic lp-functions finish correctly (FBA, FVA, yield optimization)."""
from .test_01_load_models_and_solvers import *
import straindesign as sd
from numpy import inf, isinf, isnan


def test_fba(curr_solver, model_gpr):
    """Test FBA with constraints."""
    sol = sd.fba(model_gpr, solver=curr_solver, constraints=['r3 <= 3', 'r5 = 1.5'])
    assert (round(sol.objective_value, 9) == 4.5)
    assert (sol.status == sd.OPTIMAL)


def test_fba_unbounded(curr_solver, model_gpr):
    """Test unbounded FBA."""
    for r in model_gpr.reactions:
        if r._lower_bound < 0:
            r._lower_bound = -inf
        if r._upper_bound > 0:
            r._upper_bound = inf
    sol = sd.fba(model_gpr, solver=curr_solver)
    assert (sol.status == sd.UNBOUNDED)


def test_fba_infeasible(curr_solver, model_gpr):
    """Test infeasible FBA."""
    sol = sd.fba(model_gpr, solver=curr_solver, constraints='r3 <= -2')
    assert (sol.status == sd.INFEASIBLE)


def test_fva(curr_solver, model_gpr):
    """Test FVA with constraints."""
    sol = sd.fva(model_gpr, solver=curr_solver, constraints=['r3 <= 3', 'r5 = 1.5'])
    assert (sol.shape == (11, 2))


def test_fva_infeasible(curr_solver, model_gpr):
    """Test infeasible FVA with constraints."""
    sol = sd.fva(model_gpr, solver=curr_solver, constraints=['r3 <= -3', 'r5 = 1.5'])
    assert (sol.shape == (11, 2))
    assert (isnan(sol.values[1, 1]))


def test_fva_unbounded(curr_solver, model_small_example):
    """Test FVA that is partially unbounded."""
    for r in model_small_example.reactions:
        if r.name in ['R2', 'R3', 'R9']:
            if r._lower_bound < 0:
                r._lower_bound = -inf
            if r._upper_bound > 0:
                r._upper_bound = inf
    sol = sd.fva(model_small_example, solver=curr_solver, constraints=['R5 = 1.5'])
    assert (sol.shape == (10, 2))
    assert (isinf(sol.values[1, 1]))


def test_yield_opt(curr_solver, model_weak_coupling):
    """Test yield optimization."""
    constr = ['r4 = 0', 'r7 = 0', 'r9 = 0', 'r_BM >= 4']
    num = 'r_P'
    den = 'r_S'
    sol = sd.yopt(model_weak_coupling, obj_num=num, obj_den=den, solver=curr_solver, constraints=constr)
    assert (round(sol.objective_value, 9) == 0.6)
    sol = sd.yopt(model_weak_coupling, obj_sense='min', obj_num=num, obj_den=den, solver=curr_solver, constraints=constr)
    assert (round(sol.objective_value, 9) == 0.2)


def test_yield_opt_unbounded(curr_solver, model_small_example):
    """Test yield optimization that is unbounded."""
    num = 'R4'
    den = 'R2'
    sol = sd.yopt(model_small_example, obj_num=num, obj_den=den, solver=curr_solver)
    assert (sol.status == sd.UNBOUNDED)


def test_yield_opt_infeasible(curr_solver, model_small_example):
    """Test yield optimization that is infeasible."""
    constr = ['R1 = -1']
    num = 'R4'
    den = 'R1'
    sol = sd.yopt(model_small_example, obj_num=num, obj_den=den, constraints=constr, solver=curr_solver)
    assert (sol.status == sd.INFEASIBLE)



#!/usr/bin/env python3
#
# Copyright 2022 Max Planck Insitute Magdeburg
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
#
"""Static strings used in the StrainDesign package

    Model and module

        MODEL_ID = 'model_id'

        PROTECT = 'mcs_lin'

        SUPPRESS = 'mcs_bilvl'

        SUPPRESS = 'suppress'

        PROTECT = 'protect'

        OPTKNOCK = 'optknock'

        ROBUSTKNOCK = 'robustknock'

        OPTCOUPLE = 'optcouple'
        
        MODULE_TYPE = 'module_type'
        
        CONSTRAINTS = 'constraints'
        
        INNER_OBJECTIVE = 'inner_objective'
        
        INNER_OPT_SENSE = 'inner_opt_sense'
        
        OUTER_OBJECTIVE = 'outer_objective'
        
        OUTER_OPT_SENSE = 'outer_opt_sense'
        
        PROD_ID = 'prod_id'
        
        MIN_GCP = 'min_gcp'
        
    Solvers and status codes
    
        SOLVER = 'solver'
        
        CPLEX = 'cplex'
        
        GUROBI = 'gurobi'
        
        SCIP = 'scip'
        
        GLPK = 'glpk'

        OPTIMAL = 'optimal' # from optlang interface
        
        INFEASIBLE ='infeasible' # from optlang interface
        
        TIME_LIMIT = 'time_limit' # from optlang interface
        
        UNBOUNDED = 'unbounded' # from optlang interface
        
        TIME_LIMIT_W_SOL = 'time_limit_w_sols'
        
        ERROR = 'error'
        
    Strain design setup

        KOCOST = 'ko_cost'
        
        KICOST = 'ki_cost'
        
        GKOCOST = 'gko_cost'
        
        GKICOST = 'gki_cost'
        
        REGCOST = 'reg_cost'

        MODULES = 'sd_modules'
    
        SETUP = 'sd_setup'
        
        MAX_SOLUTIONS = 'max_solutions'
        
        MAX_COST = 'max_cost'
        
        T_LIMIT = 'time_limit'
        
        SOLUTION_APPROACH = 'solution_approach'
        
        ANY = 'any'
        
        BEST = 'best'
        
        POPULATE = 'populate'
        
        SEED = 'seed'
        
    Analysis
    
        MAXIMIZE = 'maximize'
        
        MINIMIZE = 'minimize'
"""

# Model and module
MODEL_ID = 'model_id'
PROTECT = 'mcs_lin'
SUPPRESS = 'mcs_bilvl'
SUPPRESS = 'suppress'
PROTECT = 'protect'
OPTKNOCK = 'optknock'
ROBUSTKNOCK = 'robustknock'
OPTCOUPLE = 'optcouple'
MODULE_TYPE = 'module_type'
CONSTRAINTS = 'constraints'
INNER_OBJECTIVE = 'inner_objective'
INNER_OPT_SENSE = 'inner_opt_sense'
OUTER_OBJECTIVE = 'outer_objective'
OUTER_OPT_SENSE = 'outer_opt_sense'
PROD_ID = 'prod_id'
MIN_GCP = 'min_gcp'

# Solvers and status codes
SOLVER = 'solver'
CPLEX = 'cplex'
GUROBI = 'gurobi'
SCIP = 'scip'
GLPK = 'glpk'
from optlang.interface import OPTIMAL,    \
                              INFEASIBLE, \
                              TIME_LIMIT, \
                              UNBOUNDED

TIME_LIMIT_W_SOL = 'time_limit_w_sols'
ERROR = 'error'

# Strain design setup
KOCOST = 'ko_cost'
KICOST = 'ki_cost'
GKOCOST = 'gko_cost'
GKICOST = 'gki_cost'
REGCOST = 'reg_cost'
MODULES = 'sd_modules'
SETUP = 'sd_setup'
MAX_SOLUTIONS = 'max_solutions'
MAX_COST = 'max_cost'
T_LIMIT = 'time_limit'
SOLUTION_APPROACH = 'solution_approach'
ANY = 'any'
BEST = 'best'
POPULATE = 'populate'
SEED = 'seed'

# Analysis
MAXIMIZE = 'maximize'
MINIMIZE = 'minimize'



#!/usr/bin/env python3
#
# Copyright 2022 Max Planck Insitute Magdeburg
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
#
"""A collection of functions for the LP-based analysis of metabolic networks"""

from cobra.core import Solution
from cobra.util import create_stoichiometric_matrix
from cobra import Configuration
from scipy import sparse
from scipy.spatial import Delaunay  #, ConvexHull
from straindesign import MILP_LP, parse_constraints, parse_linexpr, lineqlist2mat, linexpr2dict, \
                         linexprdict2mat, SDPool, IndicatorConstraints, avail_solvers
from re import search
from straindesign.names import *
from typing import Dict, Tuple
from pandas import DataFrame
from numpy import floor, sign, mod, nan, isnan, unique, inf, isinf, full, linspace, \
                  prod, array, mean, flip, ceil, floor
from numpy.linalg import matrix_rank
from contextlib import redirect_stdout, redirect_stderr
from io import StringIO
import matplotlib.pyplot as plt
from matplotlib import use as set_matplotlib_backend
import logging

from straindesign.parse_constr import linexpr2mat, linexprdict2str


def select_solver(solver=None, model=None) -> str:
    """Select a solver for subsequent MILP/LP computations
    
    This function will determine the solver to be used for subsequend MILP/LP computations. If no
    argument is provided, this function will try to determine the currently selected solver from the
    COBRA configuration. If unavailable, the solver will be inferred from the packages available at
    package initialization and one of the solvers will be picked and retured in the prioritized 
    order: 'glpk', 'cplex', 'gurobi', 'scip'
    One may provide a solver or a model manually. This function then checks if the selected solver 
    is available, or else, if the solver indicated in the model is available. If yes, this function 
    returns the name of the solver as a str. If both arguments are specified, the function prefers
    'solver' over 'model'.
    
    Example:
        solver = select_solver('cplex')
    
    Args:
        solver (optional (str)):
        
            A user preferred solver, that should be checked for availability: 'glpk', 'cplex',
            'gurobi' or 'scip'.
            
        model (optional (cobra.Model)):
        
            A metabolic model that is an instance of the cobra.Model class. The function will try to
            dertermine the selected solver by accessing the field model.solver.
            
    Returns:
        (str):
        
            The selected solver name as a str (one of the following: 'glpk', 'cplex', 'gurobi', 'scip').
            
    """
    # first try to use selected solver
    if solver:
        if solver in avail_solvers:
            return solver
        else:
            logging.warning('Selected solver ' + solver + ' not available. Using ' + list(avail_solvers)[0] + " instead.")
    try:
        # if no solver was defined, use solver specified in model
        if hasattr(model, 'solver') and hasattr(model.solver, 'interface'):
            solver = search('(' + '|'.join(avail_solvers) + ')', model.solver.interface.__name__)
            if solver is not None:
                return solver[0]
            else:
                logging.warning('Solver specified in model (' + model.solver.interface.__name__ + ') unavailable')
        # if no solver specified in model, use solver from cobra configuration
        cobra_conf = Configuration()
        if hasattr(cobra_conf, 'solver'):
            solver = search('(' + '|'.join(avail_solvers) + ')', cobra_conf.solver.__name__)
            if solver is not None:
                return solver[0]
            else:
                logging.warning('Solver specified in cobra config (' + cobra_conf.solver.__name__ + ') unavailable')
    except:
        pass
    # if no solver is specified in cobra, fall back to list of available solvers and return the
    # first one available.
    return list(avail_solvers)[0]


def idx2c(i, prev) -> 'list':
    """Helper function for parallel FVA
    
    Builds the objective function for minimizing or maximizing the flux through the reaction
    with the index floor(i / 2). If i is even, there is a maximization.
    
    Args:
        i (float):
            An index between 0 and 2*num_reacs.
        prev (optional (str)):
            Index of the previously optimized reaction.
    Returns:
        (list):
            An optimization vector.
    """
    col = int(floor(i / 2))
    sig = sign(mod(i, 2) - 0.5)
    C = [[col, sig], [prev, 0.0]]
    C_idx = [C[i][0] for i in range(len(C))]
    C_idx = unique([C_idx.index(C_idx[i]) for i in range(len(C_idx))])
    C = [C[i] for i in C_idx]
    return C


def fva_worker_init(A_ineq, b_ineq, A_eq, b_eq, lb, ub, solver):
    """Helper function for parallel FVA
    
    Initialize the LP that will be solved iteratively. Is executed on workers, not on main thread.
    
    Args:
        A_ineq, b_ineq, A_eq, b_eq, lb, ub:
            The LP.
        solver (str):
            Solver to be used.
    """
    global lp_glob
    # redirect output to empty stream. Perhaps avoids some multithreading issues
    with redirect_stdout(StringIO()), redirect_stderr(StringIO()):
        lp_glob = MILP_LP(A_ineq=A_ineq, b_ineq=b_ineq, A_eq=A_eq, b_eq=b_eq, lb=lb, ub=ub, solver=solver)
        if lp_glob.solver == 'cplex':
            lp_glob.backend.parameters.threads.set(1)
            #lp_glob.backend.parameters.lpmethod.set(1)
        lp_glob.prev = 0


def fva_worker_compute(i) -> Tuple[int, float]:
    """Helper function for parallel FVA
    
    Run a single LP as a step of FVA. Is executed on workers, not on main thread.
    
    Args:
        i (int):
            Index of the computation step.
    """
    global lp_glob
    with redirect_stdout(StringIO()), redirect_stderr(StringIO()):
        C = idx2c(i, lp_glob.prev)
        if lp_glob.solver in ['cplex', 'gurobi']:
            lp_glob.backend.set_objective_idx(C)
            min_cx = lp_glob.backend.slim_solve()
        else:
            lp_glob.set_objective_idx(C)
            min_cx = lp_glob.slim_solve()
        lp_glob.prev = C[0][0]
        return i, min_cx


# GLPK needs a workaround, because problems cannot be solved in a different thread
# which apparently happens with the multiprocess


def fva_worker_init_glpk(A_ineq, b_ineq, A_eq, b_eq, lb, ub):
    """Helper function for parallel FVA
    
    Initialize the LP for GLPK that will be solved iteratively. Is executed on workers, not on main thread.
    
    Args:
        A_ineq, b_ineq, A_eq, b_eq, lb, ub:
            The LP.
    """
    global lp_glob
    lp_glob = {}
    lp_glob['A_ineq'] = A_ineq
    lp_glob['b_ineq'] = b_ineq
    lp_glob['A_eq'] = A_eq
    lp_glob['b_eq'] = b_eq
    lp_glob['lb'] = lb
    lp_glob['ub'] = ub


def fva_worker_compute_glpk(i) -> Tuple[int, float]:
    """Helper function for parallel FVA
    
    Run a single LP for GLPK as a step of FVA. Is executed on workers, not on main thread.
    
    Args:
        i (int):
            Index of the computation step.
    """
    global lp_glob
    with redirect_stdout(StringIO()), redirect_stderr(StringIO()):
        lp_i = MILP_LP(A_ineq=lp_glob['A_ineq'],
                       b_ineq=lp_glob['b_ineq'],
                       A_eq=lp_glob['A_eq'],
                       b_eq=lp_glob['b_eq'],
                       lb=lp_glob['lb'],
                       ub=lp_glob['ub'],
                       solver=GLPK)
        col = int(floor(i / 2))
        sig = sign(mod(i, 2) - 0.5)
        lp_i.set_objective_idx([[col, sig]])
        min_cx = lp_i.slim_solve()
    return i, min_cx


def fva(model, **kwargs) -> DataFrame:
    """Flux Variability Analysis (FVA)
    
    Flux Variability Analysis determines the global flux ranges of reactions by minimizing and 
    maximizing the flux through all reactions of a given metabolic network. This FVA function
    additionally allows the user to narrow down the flux states with additional constraints.
    
    Example:
        flux_ranges = fva(model, constraints='EX_o2_e=0', solver='gurobi')
    
    Args:
        model (cobra.Model):
            A metabolic model that is an instance of the cobra.Model class.
            
        solver (optional (str)):
            The solver that should be used for FVA.
            
        constraints (optional (str) or (list of str) or (list of [dict,str,float])): (Default: '')
            List of *linear* constraints to be applied on top of the model: signs + or -, scalar 
            factors for reaction rates, inclusive (in)equalities and a float value on the right hand 
            side. The parsing of the constraints input allows for some flexibility. Correct (and 
            identical) inputs are, for instance: 
            constraints='-EX_o2_e <= 5, ATPM = 20' or
            constraints=['-EX_o2_e <= 5', 'ATPM = 20'] or
            constraints=[[{'EX_o2_e':-1},'<=',5], [{'ATPM':1},'=',20]]
            
    Returns:
        (pandas.DataFrame):
            A data frame containing the minimum and maximum attainable flux rates for all reactions.
    """
    reaction_ids = model.reactions.list_attr("id")
    numr = len(model.reactions)

    if CONSTRAINTS in kwargs and kwargs[CONSTRAINTS]:
        kwargs[CONSTRAINTS] = parse_constraints(kwargs[CONSTRAINTS], reaction_ids)
        A_ineq, b_ineq, A_eq, b_eq = lineqlist2mat(kwargs[CONSTRAINTS], reaction_ids)

    if SOLVER not in kwargs:
        kwargs[SOLVER] = None
    solver = select_solver(kwargs[SOLVER], model)

    # prepare vectors and matrices
    A_eq_base = sparse.csr_matrix(create_stoichiometric_matrix(model))
    b_eq_base = [0] * len(model.metabolites)
    if 'A_eq' in locals():
        A_eq = sparse.vstack((A_eq_base, A_eq))
        b_eq = b_eq_base + b_eq
    else:
        A_eq = A_eq_base
        b_eq = b_eq_base
    if 'A_ineq' not in locals():
        A_ineq = sparse.csr_matrix((0, numr))
        b_ineq = []
    lb = [v.lower_bound for v in model.reactions]
    ub = [v.upper_bound for v in model.reactions]

    # build LP
    lp = MILP_LP(A_ineq=A_ineq, b_ineq=b_ineq, A_eq=A_eq, b_eq=b_eq, lb=lb, ub=ub, solver=solver)
    _, _, status = lp.solve()
    if status not in [OPTIMAL, UNBOUNDED]:  # if problem not feasible or unbounded
        logging.error('FVA problem not feasible.')
        return DataFrame(
            {
                "minimum": [nan for i in range(1, 2 * numr, 2)],
                "maximum": [nan for i in range(0, 2 * numr, 2)],
            },
            index=reaction_ids,
        )

    processes = Configuration().processes
    num_reactions = len(reaction_ids)
    processes = min(processes, num_reactions)

    x = [nan] * 2 * numr

    # Dummy to check if optimization runs
    # worker_init(A_ineq,b_ineq,A_eq,b_eq,lb,ub,solver)
    # worker_compute(1)
    if processes > 1 and numr > 300:  #and solver != 'GLPK': # activate alternative routine with GLPK, if issues arise
        # with Pool(processes,initializer=worker_init,initargs=(A_ineq,b_ineq,A_eq,b_eq,lb,ub,solver)) as pool:
        with SDPool(processes, initializer=fva_worker_init, initargs=(A_ineq, b_ineq, A_eq, b_eq, lb, ub, solver)) as pool:
            chunk_size = len(reaction_ids) // processes
            # x = pool.imap_unordered(worker_compute, range(2*numr), chunksize=chunk_size)
            for i, value in pool.imap_unordered(fva_worker_compute, range(2 * numr), chunksize=chunk_size):
                x[i] = value
    # GLPK works better when reinitializing the LP in every iteration. Unfortunately, this is slow
    # but for now by far the most stable solution.
    elif processes > 1 and numr > 500 and solver == GLPK:
        with SDPool(processes, initializer=fva_worker_init_glpk, initargs=(A_ineq, b_ineq, A_eq, b_eq, lb, ub)) as pool:
            chunk_size = len(reaction_ids) // processes
            # # x = pool.imap_unordered(worker_compute, range(2*numr), chunksize=chunk_size)
            for i, value in pool.imap_unordered(fva_worker_compute_glpk, range(2 * numr), chunksize=chunk_size):
                x[i] = value
    else:
        fva_worker_init(A_ineq, b_ineq, A_eq, b_eq, lb, ub, solver)
        for i in range(2 * numr):
            _, x[i] = fva_worker_compute(i)

    x = [v if abs(v) >= 1e-11 else 0.0 for v in x]  # cut off for very small absolute values
    fva_result = DataFrame(
        {
            "minimum": [x[i] for i in range(1, 2 * numr, 2)],
            "maximum": [-x[i] for i in range(0, 2 * numr, 2)],
        },
        index=reaction_ids,
    )

    return fva_result


def fba(model, **kwargs) -> Solution:
    """Flux Balance Analysis (FBA), parsimonius Flux Balance Analysis (pFBA),
    
    Flux Balance Analysis optimizes a *linear objective function* in a space of steady-state
    flux vectors given by a constraint-based metabolic model. FVA is often used to determine
    the (stoichiometrically) maximal possible growth rate, or flux rate towards a particular
    product. This FBA function allows to us a custom objective function and sense and 
    allows the user to narrow down the flux states with additional constraints. In addition, 
    one may use different types of parsimonious FBAs to either reduce the total sum of fluxes
    or the total number of active reactions after the primary objective is optimized.    
    
    Example:
        optim = fba(model, constraints='EX_o2_e=0', solver='gurobi', pfba=1)
    
    Args:
        model (cobra.Model):
            A metabolic model that is an instance of the cobra.Model class. If no custom objective
            function is provided, the model's objective function is retrieved from the fields
            model.reactions[i].objective_coefficient.
            
        solver (optional (str)):
            The solver that should be used for FBA.
            
        constraints (optional (str) or (list of str) or (list of [dict,str,float])): (Default: '')
            List of *linear* constraints to be applied on top of the model: signs + or -, scalar 
            factors for reaction rates, inclusive (in)equalities and a float value on the right hand 
            side. The parsing of the constraints input allows for some flexibility. Correct (and 
            identical) inputs are, for instance: 
            constraints='-EX_o2_e <= 5, ATPM = 20' or
            constraints=['-EX_o2_e <= 5', 'ATPM = 20'] or
            constraints=[[{'EX_o2_e':-1},'<=',5], [{'ATPM':1},'=',20]]
            
        obj (optional (str) or (dict)):
            As a custom objective function, any linear expression can be used, either provided as a 
            single string or as a dict. Correct (and identical) inputs are, for instance:
            inner_objective='BIOMASS_Ecoli_core_w_GAM'
            inner_objective={'BIOMASS_Ecoli_core_w_GAM': 1}
            
        obj_sense (optional (str)): (Default: 'maximize')
            The optimization direction can be set either to 'maximize' (or 'max') or 'minimize' (or 'min').
            
        pfba (optional (int)): (Default: 0)
            The level of parsimonious FBA that should be applied. 0: no pFBA, only optmize the primary 
            objective, 1: minimize sum of fluxes after the primary objective is optimized, 2: minimize the
            number of active reactions after the primary objective is optimized.
            
    Returns:
        (cobra.core.Solution):
            A solution object that contains the objective value, an optimal flux vector and the optmization
            status.
    """
    reaction_ids = model.reactions.list_attr("id")

    if CONSTRAINTS in kwargs:
        kwargs[CONSTRAINTS] = parse_constraints(kwargs[CONSTRAINTS], reaction_ids)
        A_ineq, b_ineq, A_eq, b_eq = lineqlist2mat(kwargs[CONSTRAINTS], reaction_ids)
    else:
        kwargs[CONSTRAINTS] = []

    if 'obj' in kwargs and kwargs['obj'] is not None:
        if type(kwargs['obj']) is str:
            kwargs['obj'] = linexpr2dict(kwargs['obj'], reaction_ids)
        c = linexprdict2mat(kwargs['obj'], reaction_ids).toarray()[0].tolist()
    else:
        c = [i.objective_coefficient for i in model.reactions]

    if ('obj_sense' not in kwargs and model.objective_direction == 'max') or \
       ('obj_sense' in kwargs and kwargs['obj_sense'] not in ['min','minimize']):
        obj_sense = 'maximize'
        c = [-i for i in c]
    else:
        obj_sense = 'minimize'

    if 'pfba' in kwargs:
        pfba = kwargs['pfba']
    else:
        pfba = False

    if SOLVER not in kwargs:
        kwargs[SOLVER] = None
    solver = select_solver(kwargs[SOLVER], model)

    # prepare vectors and matrices
    A_eq_base = create_stoichiometric_matrix(model)
    A_eq_base = sparse.csr_matrix(A_eq_base)
    b_eq_base = [0.0] * len(model.metabolites)
    if 'A_eq' in locals():
        A_eq = sparse.vstack((A_eq_base, A_eq))
        b_eq = b_eq_base + b_eq
    else:
        A_eq = A_eq_base
        b_eq = b_eq_base
    if 'A_ineq' not in locals():
        A_ineq = sparse.csr_matrix((0, len(model.reactions)))
        b_ineq = []
    lb = [v.lower_bound for v in model.reactions]
    ub = [v.upper_bound for v in model.reactions]

    # build LP
    fba_prob = MILP_LP(c=c, A_ineq=A_ineq, b_ineq=b_ineq, A_eq=A_eq, b_eq=b_eq, lb=lb, ub=ub, solver=solver)

    x, opt_cx, status = fba_prob.solve()
    if obj_sense == 'minimize':
        opt_cx = -opt_cx
    if status == UNBOUNDED:
        num_prob = MILP_LP(c=[-v for v in c], A_ineq=A_ineq, b_ineq=b_ineq, A_eq=A_eq, b_eq=b_eq, lb=lb, ub=ub, solver=solver)
        min_cx = num_prob.slim_solve()
        if min_cx <= 0 or isnan(min_cx):
            num_prob.add_eq_constraints(c, [-1.0])
        else:
            num_prob.add_eq_constraints(c, min_cx)
        x, _, _ = num_prob.solve()
    elif status not in [OPTIMAL, UNBOUNDED]:
        status = INFEASIBLE
    if pfba and status == OPTIMAL:  # for pfba, split all reversible reactions and minimize the total flux
        numr = len(c)
        if pfba == 2:  # pfba mode 2 minimizes the number of active reactions
            # fix optimal flux and do fva to get essential reactions and speed up minimization
            kwargs_fva = kwargs.copy()
            kwargs_fva[CONSTRAINTS].append([{reaction_ids[i]: c[i] for i in range(numr) if c[i] != 0}, '=', opt_cx])
            fva_sol = fva(model, **kwargs_fva)

            A_ineq_pfba2 = A_ineq.copy()
            A_ineq_pfba2.resize(A_ineq.shape[0], 2 * numr)
            A_eq_pfba2 = A_eq.copy()
            A_eq_pfba2.resize(A_eq.shape[0], 2 * numr)
            lb_pfba2 = lb + [0.0] * numr
            # only set non-essential reactions as knockable
            ub_pfba2 = ub + [0.0 if prod(sign(lim)) > 0 else 1.0 for _, lim in fva_sol.iterrows()]
            c_pfba2 = [0.0] * numr + [-1.0] * numr
            A_ic = sparse.csr_matrix(([1] * numr, ([j for j in range(numr)], [j for j in range(numr)])), [numr, 2 * numr])
            ic = IndicatorConstraints([numr + j for j in range(numr)], A_ic, [0] * numr, 'E' * numr, [1.0] * numr)
            pfba2_prob = MILP_LP(c=c_pfba2,
                                 A_ineq=A_ineq_pfba2,
                                 b_ineq=b_ineq,
                                 A_eq=A_eq_pfba2,
                                 b_eq=b_eq,
                                 lb=lb_pfba2,
                                 ub=ub_pfba2,
                                 indic_constr=ic,
                                 vtype='C' * numr + 'B' * numr,
                                 solver=solver)
            pfba2_prob.add_eq_constraints(c + [0] * numr, [opt_cx])
            y, _, _ = pfba2_prob.solve()
            zero_flux = [i for i, j in enumerate(range(numr, 2 * numr)) if y[j]]
            lb = [l if i not in zero_flux else 0.0 for i, l in enumerate(lb)]
            ub = [u if i not in zero_flux else 0.0 for i, u in enumerate(ub)]
        A_ineq_pfba = sparse.hstack((A_ineq, -A_ineq))
        A_eq_pfba = sparse.hstack((A_eq, -A_eq))
        lb_pfba = [max((0, l)) for l in lb] + [max((0, -u)) for u in ub]
        ub_pfba = [max((0, u)) for u in ub] + [max((0, -l)) for l in lb]
        c_pfba = c + [-v for v in c]
        pfba_prob = MILP_LP(c=[1.0] * 2 * numr,
                            A_ineq=A_ineq_pfba,
                            b_ineq=b_ineq,
                            A_eq=A_eq_pfba,
                            b_eq=b_eq,
                            lb=lb_pfba,
                            ub=ub_pfba,
                            solver=solver)
        pfba_prob.add_eq_constraints(c_pfba, [opt_cx])
        x, _, _ = pfba_prob.solve()
        x = [x[i] - x[j] for i, j in enumerate(range(numr, 2 * numr))]

    x = [v if abs(v) >= 1e-11 else 0.0 for v in x]  # cut off for very small absolute values
    fluxes = {reaction_ids[i]: x[i] for i in range(len(x))}
    sol = Solution(objective_value=-opt_cx, status=status, fluxes=fluxes)
    return sol


def yopt(model, **kwargs) -> Solution:
    """Yield optmization (YOpt)
    
    Yield optimization optimizes a *fractional objective function* in a space of steady-state
    flux vectors given by a constraint-based metabolic model. Yield optimization employs linear
    fractional programming, and is often utilized to determine the (stoichiometrically) maximal 
    possible product yield, that is, the fraction between the product exchange and the substrate
    uptake flux. This function requires a custom fractional objective function specified by a 
    *linear* numerator and denominator terms. Coefficients in the linear numerator or denominator
    expression can be used to optimize for carbon recovery, for instance: 
    objective: (3*pyruvate_ex)/(2*ac_up+6*glc_up)
    The user may also specify the optimization sense. In addition, additional constraints can be 
    specified to narrow down the flux space.
    
    Yield optimization can fail because of several reasons. Here is how the function reacts:
    
    1. The model is infeasible:
        The function returns infeasible with no flux vector
    2. The denominator is fixed to zero:
        The function returns infeasible with no flux vector
    3. The numerator is unbounded:
        The function returns unbounded with no flux vector
    4. The denominator can become zero:
        The function returns unbounded, and a flux vector is computed by fixing the
        the denominator
        
    Example:
        optim = yopt(model, obj_num='2 EX_etoh_e', obj_den='-6 EX_glc__D_e', constraints='EX_o2_e=0')
    
    Args:
        model (cobra.Model):
            A metabolic model that is an instance of the cobra.Model class.
            
        obj_num ((str) or (dict)):
            The numerator of the fractional objective function, provided as a linear expression,
            either as a character string or as a dict. E.g.: obj_num='EX_prod_e' or 
            obj_num={'EX_prod_e': 1}
            
        obj_den ((str) or (dict)):
            The denominator of the fractional objective function, provided as a linear expression,
            either as a character string or as a dict. E.g.: obj_num='1.0 EX_subst_e' or 
            obj_num={'EX_subst_e': 1}
            
        obj_sense (optional (str)): (Default: 'maximize')
            The optimization direction can be set either to 'maximize' (or 'max') or 'minimize' (or 'min').
            
        solver (optional (str)):
            The solver that should be used for YOpt.
            
        constraints (optional (str) or (list of str) or (list of [dict,str,float])): (Default: '')
            List of *linear* constraints to be applied on top of the model: signs + or -, scalar 
            factors for reaction rates, inclusive (in)equalities and a float value on the right hand 
            side. The parsing of the constraints input allows for some flexibility. Correct (and 
            identical) inputs are, for instance: 
            constraints='-EX_o2_e <= 5, ATPM = 20' or
            constraints=['-EX_o2_e <= 5', 'ATPM = 20'] or
            constraints=[[{'EX_o2_e':-1},'<=',5], [{'ATPM':1},'=',20]]

    Returns:
        (cobra.core.Solution):
            A solution object that contains the objective value, an optimal flux vector and the optmization
            status.
    """
    reaction_ids = model.reactions.list_attr("id")
    if 'obj_num' not in kwargs:
        raise Exception('For a yield optimization, the numerator expression must be provided under the keyword "obj_num".')
    else:
        if type(kwargs['obj_num']) is not dict:
            obj_num = linexpr2mat(kwargs['obj_num'], reaction_ids)
        else:
            obj_num = linexprdict2mat(kwargs['obj_num'], reaction_ids)

    if 'obj_den' not in kwargs:
        raise Exception('For a yield optimization, the denominator expression must be provided under the keyword "obj_den".')
    else:
        if type(kwargs['obj_den']) is not dict:
            obj_den = linexpr2mat(kwargs['obj_den'], reaction_ids)
        else:
            obj_den = linexprdict2mat(kwargs['obj_den'], reaction_ids)

    if 'obj_sense' not in kwargs or kwargs['obj_sense'] not in ['min', 'minimize']:
        obj_sense = 'maximize'
    else:
        obj_sense = 'minimize'
        obj_num = -obj_num

    if CONSTRAINTS in kwargs:
        kwargs[CONSTRAINTS] = parse_constraints(kwargs[CONSTRAINTS], reaction_ids)
        A_ineq, b_ineq, A_eq, b_eq = lineqlist2mat(kwargs[CONSTRAINTS], reaction_ids)

    if SOLVER not in kwargs:
        kwargs[SOLVER] = None
    solver = select_solver(kwargs[SOLVER], model)

    # prepare vectors and matrices for base problem
    A_eq_base = create_stoichiometric_matrix(model)
    A_eq_base = sparse.csr_matrix(A_eq_base)
    b_eq_base = [0] * len(model.metabolites)
    if 'A_eq' in locals():
        A_eq = sparse.vstack((A_eq_base, A_eq), 'csr')
        b_eq = b_eq_base + b_eq
    else:
        A_eq = A_eq_base
        b_eq = b_eq_base
    if 'A_ineq' not in locals():
        A_ineq = sparse.csr_matrix((0, len(model.reactions)))
        b_ineq = []
    # Integrate upper and lower bounds into A_ineq and b_ineq
    real_lb = [i for i, v in enumerate(model.reactions.list_attr('lower_bound')) if not isinf(v)]
    real_ub = [i for i, v in enumerate(model.reactions.list_attr('upper_bound')) if not isinf(v)]
    sparse_lb = sparse.coo_matrix(([-1] * len(real_lb), (range(len(real_lb)), real_lb)), (len(real_lb), A_ineq.shape[1]))
    sparse_ub = sparse.coo_matrix(([1] * len(real_ub), (range(len(real_ub)), real_ub)), (len(real_ub), A_ineq.shape[1]))
    A_ineq = sparse.vstack((A_ineq, sparse_lb, sparse_ub), 'csr')
    b_ineq = b_ineq + [-model.reactions[i].lower_bound for i in real_lb] + \
                      [ model.reactions[i].upper_bound for i in real_ub]
    # Analyze maximum and minimum value of denominator function to decide whether to fix it to +1 or -1 or abort computation
    den_sign = []
    den_prob = MILP_LP(c=obj_den.todense().tolist()[0], A_ineq=A_ineq, b_ineq=b_ineq, A_eq=A_eq, b_eq=b_eq, solver=solver)
    _, min_denx, status_i = den_prob.solve()
    # check model feasibility
    if status_i not in [OPTIMAL, UNBOUNDED]:
        fluxes = {reaction_ids[i]: nan for i in range(len(reaction_ids))}
        return Solution(objective_value=nan, status=INFEASIBLE, fluxes=fluxes)
    # is minimum of denominator term negative?
    if min_denx < 0:
        den_sign += [-1]
    # is maximum of denominator term positive?
    den_prob.set_objective((-obj_den).todense().tolist()[0])
    _, max_denx, _ = den_prob.solve()
    if max_denx < 0:
        den_sign += [1]
    # is denominator fixed to zero
    if not den_sign:
        logging.error('Denominator term can only take the value 0. Yield computation impossible.')
        fluxes = {reaction_ids[i]: nan for i in range(len(reaction_ids))}
        return Solution(objective_value=nan, status=INFEASIBLE, fluxes=fluxes)

    # Create linear fractional problem (LFP)
    # A variable is added here to scale the right hand side of the original problem
    A_ineq_lfp = sparse.hstack((A_ineq, sparse.csr_matrix([-b for b in b_ineq]).transpose()), 'csr')
    b_ineq_lfp = [0 for _ in b_ineq]
    A_eq_lfp = sparse.vstack((  sparse.hstack((A_eq,sparse.csr_matrix([-b for b in b_eq]).transpose())),\
                            sparse.hstack((obj_den,sparse.csr_matrix((1, 1))))),'csr')
    opt_cx = inf
    for d in den_sign:
        b_eq_lfp = [0 for _ in b_eq] + [d]
        # build LP
        yopt_prob = MILP_LP(c=(-d * obj_num).todense().tolist()[0] + [0],
                            A_ineq=A_ineq_lfp,
                            b_ineq=b_ineq_lfp,
                            A_eq=A_eq_lfp,
                            b_eq=b_eq_lfp,
                            solver=solver)
        x_i, opt_i, status_i = yopt_prob.solve()
        if opt_i < opt_cx:
            x = x_i
            opt_cx = opt_i
            status = status_i
    if status is OPTIMAL:
        factor = x[-1]  # get factor from LFP
        if factor == 0:
            factor = 1
        fluxes = {r: x[i] / factor for i, r in enumerate(reaction_ids)}
        if obj_sense == 'maximize':
            opt_cx = -opt_cx  # correct sign (maximization)
        sol = Solution(objective_value=opt_cx, status=status, fluxes=fluxes)
        if x[-1] == 0:
            sol.scalable = True
            logging.info('Solution flux vector may be scaled with an arbitrary factor.')
        else:
            sol.scalable = False
        return sol
    elif status is UNBOUNDED:
        opt_cx = nan
        # check if numerator can be nonzero when denominator is zero
        fct = 1 - 2 * (obj_sense == 'maximize')
        den_prob.set_objective((fct * obj_num).todense().tolist()[0])
        den_prob.add_eq_constraints(obj_den, [0])
        x, max_num, status_i = den_prob.solve()
        if isinf(max_num) or status_i == INFEASIBLE:  # if numerator is still unbounded, generate a fixed solution
            num_prob = MILP_LP(c=(-fct * obj_num).todense().tolist()[0], A_ineq=A_ineq, b_ineq=b_ineq, A_eq=A_eq, b_eq=b_eq, solver=solver)
            min_num = num_prob.slim_solve()
            if min_num <= 0:
                num_prob.add_eq_constraints((-fct * obj_num).todense().tolist()[0], [1])
            else:
                num_prob.add_eq_constraints((-fct * obj_num).todense().tolist()[0], min_num)
            x, opt_i, status_i = num_prob.solve()
        if not isnan(max_num):
            if max_num == 0:
                num_prob = MILP_LP(c=(-fct * obj_num).todense().tolist()[0],
                                   A_ineq=A_ineq,
                                   b_ineq=b_ineq,
                                   A_eq=A_eq,
                                   b_eq=b_eq,
                                   solver=solver)
                num_prob.add_eq_constraints((obj_den).todense().tolist()[0], [0])
                x, opt_i, status_i = num_prob.solve()
                fluxes = {r: x[i] for i, r in enumerate(reaction_ids)}
                logging.warning('Yield is undefined because denominator can become zero. Solution '\
                    'flux vector maximizes the numerator.')
                sol = Solution(objective_value=opt_cx, status=status, fluxes=fluxes)
                sol.scalable = False
            else:
                fluxes = {r: x[i] for i, r in enumerate(reaction_ids)}
                logging.warning('Yield is undefined because denominator can become zero. Solution '\
                    'flux vector maximizes the numerator.')
                sol = Solution(objective_value=opt_cx, status=status, fluxes=fluxes)
                sol.scalable = False
            return sol
        else:
            if obj_sense == 'maximize':
                opt_cx = inf
            else:
                opt_cx = -inf
            fluxes = {r: x[i] for i, r in enumerate(reaction_ids)}
            logging.warning('Yield is infinite because the numerator is unbounded.')
            sol = Solution(objective_value=opt_cx, status=status, fluxes=fluxes)
            sol.scalable = True
            return sol
    else:
        status = INFEASIBLE


def plot_flux_space(model, axes, **kwargs) -> Tuple[list, list, list]:
    """Plot projections of the space of steady-state flux vectors onto two or three dimensions.
    
    This function uses LP and matplotlib to generate lower dimensional representations of the 
    flux space. Custom *linear* or *fractional-linear* expressions can be used for the plot
    axis. The most commonly used flux space reprentations are the *production envelope* that
    plots the growth rate (x) vs the product synthesis rate (y) and the *yield space plot*
    that plots the biomass yield (x) vs the product yiel (y). One may specify additional
    constraints to investigate subspaces of the metabolic steady-state flux space.
    
    Example:
        plot_flux_space(model,('BIOMASS_Ecoli_core_w_GAM','EX_etoh_e'))
    
    Args:
        model (cobra.Model):
            A metabolic model that is an instance of the cobra.Model class.
            
        axes ((list of lists) or (list of str)):
            A set of linear expressions that specify which reactions/expressions/dimensions should be
            used on the axis. Examples: axes=['BIOMASS_Ecoli_core_w_GAM','EX_etoh_e'] or 
            axes=[['BIOMASS_Ecoli_core_w_GAM','-EX_glc_e'],['EX_etoh_e','-EX_glc_e']] or
            axes=[['BIOMASS_Ecoli_core_w_GAM'],['EX_etoh_e','-EX_glc_e']]
            
        solver (optional (str)):
            The solver that should be used for scanning the flux space.
            
        constraints (optional (str) or (list of str) or (list of [dict,str,float])): (Default: '')
            List of *linear* constraints to be applied on top of the model: signs + or -, scalar 
            factors for reaction rates, inclusive (in)equalities and a float value on the right hand 
            side. The parsing of the constraints input allows for some flexibility. Correct (and 
            identical) inputs are, for instance: 
            constraints='-EX_o2_e <= 5, ATPM = 20' or
            constraints=['-EX_o2_e <= 5', 'ATPM = 20'] or
            constraints=[[{'EX_o2_e':-1},'<=',5], [{'ATPM':1},'=',20]]
            
        plt_backend (optional (str)):
            The matplotlib backend that should be used for plotting:
            interactive backends: GTK3Agg, GTK3Cairo, GTK4Agg, GTK4Cairo, MacOSX, nbAgg, QtAgg, QtCairo,
                                  TkAgg, TkCairo, WebAgg, WX, WXAgg, WXCairo, Qt5Agg, Qt5Cairo
            non-interactive backends: agg, cairo, pdf, pgf, ps, svg, template
            
        show (optional (bool)): (Default: True)
            Should matplotlib show the plot or should it stop after plot generation. show=False can
            be useful if flux spaces should be plotted and saved in a non-interactive environment, 
            multiple flux spaces should be plotted at once or the plot should be modified before been shown.
        
        points (optional (int)): (Default: 25 (3D) or 40 (2D))
            The number of intervals in which the flux space should be sampled along each axis. A higher
            number will increase resoltion but also computation time.

    Returns:
        (Tuple):
            (datapoints, triang, plot1). The array of datapoints from which the plot was generated. These
            datapoints are optimal values for different optimizations within the flux space. The triang
            variable contains information about which datapoints need to be connected in triangles to
            render a consistend 3-D surface with Delaunay triangles (Delaunay triangulation). The last
            variable contains the matplotlib object.
    """
    reaction_ids = model.reactions.list_attr("id")

    if CONSTRAINTS in kwargs:
        kwargs[CONSTRAINTS] = parse_constraints(kwargs[CONSTRAINTS], reaction_ids)
    else:
        kwargs[CONSTRAINTS] = []

    if SOLVER not in kwargs:
        kwargs[SOLVER] = None
    solver = select_solver(kwargs[SOLVER], model)

    if 'plt_backend' in kwargs:
        set_matplotlib_backend(kwargs['plt_backend'])

    if 'show' not in kwargs:
        show = True
    else:
        show = kwargs['show']

    axes = [list(ax) if not isinstance(ax, str) else [ax] for ax in axes]  # cast to list of lists
    num_axes = len(axes)
    if num_axes not in [2, 3]:
        raise Exception('Please define 2 or 3 axes as a list of tuples [ax1, ax2, (optional) ax3] with ax1 = (den,num).\n'+\
                        '"den" and "num" being linear expressions.')

    if 'points' in kwargs:
        points = kwargs['points']
    else:
        if num_axes == 2:
            points = 40
        else:
            points = 25

    ax_name = ["" for _ in range(num_axes)]
    ax_limits = [(nan, nan) for _ in range(num_axes)]
    val_limits = [(nan, nan) for _ in range(num_axes)]
    ax_type = ["" for _ in range(num_axes)]
    for i, ax in enumerate(axes):
        if len(ax) == 1:
            ax_type[i] = 'rate'
        elif len(ax) == 2:
            ax_type[i] = 'yield'
        if ax_type[i] == 'rate':
            ax[0] = parse_linexpr(ax[0], reaction_ids)[0]
            ax_name[i] = linexprdict2str(ax[0])
            sol_min = fba(model, obj=ax[0], constraints=kwargs[CONSTRAINTS], solver=solver, obj_sense='minimize')
            sol_max = fba(model, obj=ax[0], constraints=kwargs[CONSTRAINTS], solver=solver, obj_sense='maximize')
            # abort if any of the fluxes are unbounded or undefined
            inval = [i + 1 for i, v in enumerate([sol_min, sol_max]) if v.status == UNBOUNDED or v.status == INFEASIBLE]
            if any(inval):
                raise Exception('One of the specified reactions is unbounded or problem is infeasible. Plot cannot be generated.')
        elif ax_type[i] == 'yield':
            ax[0] = parse_linexpr(ax[0], reaction_ids)[0]
            ax[1] = parse_linexpr(ax[1], reaction_ids)[0]
            ax_name[i] = '(' + linexprdict2str(ax[0]) + ') / (' + linexprdict2str(ax[1]) + ')'
            sol_min = yopt(model, obj_num=ax[0], obj_den=ax[1], constraints=kwargs[CONSTRAINTS], solver=solver, obj_sense='minimize')
            sol_max = yopt(model, obj_num=ax[0], obj_den=ax[1], constraints=kwargs[CONSTRAINTS], solver=solver, obj_sense='maximize')
            # abort if any of the yields are unbounded or undefined
            inval = [i + 1 for i, v in enumerate([sol_min, sol_max]) if v.status == UNBOUNDED or v.status == INFEASIBLE]
            if any(inval):
                raise Exception('One of the specified yields is unbounded or undefined or problem is infeasible. Plot cannot be generated.')
        val_limits[i] = [ceil_dec(sol_min.objective_value, 8), floor_dec(sol_max.objective_value, 8)]
        ax_limits[i] = [min((0, val_limits[i][0])), max((0, val_limits[i][1]))]

    # compute points
    x_space = linspace(val_limits[0][0], val_limits[0][1], num=points).tolist()
    lb = full(points, nan)
    ub = full(points, nan)
    for i, x in enumerate(x_space):
        constr = kwargs[CONSTRAINTS].copy()
        if ax_type[0] == 'rate':
            constr += [[axes[0][0], '=', x]]
        elif ax_type[0] == 'yield':
            constr += [[{**axes[0][0], **{k: -v * x for k, v in axes[0][1].items()}}, '=', 0]]
        if ax_type[1] == 'rate':
            sol_vmin = fba(model, constraints=constr, obj=axes[1][0], solver=solver, obj_sense='minimize')
            sol_vmax = fba(model, constraints=constr, obj=axes[1][0], solver=solver, obj_sense='maximize')
        elif ax_type[1] == 'yield':
            sol_vmin = yopt(model, constraints=constr, obj_num=axes[1][0], obj_den=axes[1][1], solver=solver, obj_sense='minimize')
            sol_vmax = yopt(model, constraints=constr, obj_num=axes[1][0], obj_den=axes[1][1], solver=solver, obj_sense='maximize')
        lb[i] = ceil_dec(sol_vmin.objective_value, 9)
        ub[i] = floor_dec(sol_vmax.objective_value, 9)

    if num_axes == 2:
        datapoints = [[x, l] for x, l in zip(x_space, lb)] + [[x, u] for x, u in zip(x_space, ub)]
        datapoints_bottom = [i for i, _ in enumerate(x_space)]
        datapoints_top = [len(x_space) + i for i, _ in enumerate(x_space)]
        # triangles 2D (only for return)
        triang = []
        for i in range(len(x_space) - 1):
            temp_points = [datapoints_bottom[i], datapoints_top[i], datapoints_bottom[i + 1], datapoints_top[i + 1]]
            pts = [array(datapoints[idx_p]) for idx_p in temp_points]
            try:
                if matrix_rank(pts - pts[0]) > 1:
                    triang_temp = Delaunay(pts).simplices
                    triang += [[temp_points[idx] for idx in p] for p in triang_temp]
            except:
                logging.warning('Computing matrix rank or Delaunay simplices failed.')
        # Plot
        x = [v for v in x_space] + [v for v in reversed(x_space)]
        y = [v for v in lb] + [v for v in reversed(ub)]
        if lb[0] != ub[0]:
            x.extend([x_space[0], x_space[0]])
            y.extend([lb[0], ub[0]])
        plot1 = plt.fill(x, y, linewidth=0.5)  # edgecolor='mediumblue',
        plot1 = plot1[0]
        # plot1 = plt.plot(x, y)

        # Alternatively, this can be plotted as triangles
        # One may use this code-snippet to plot the return value of this function
        # x = [d[0] for d in datapoints]
        # y = [d[1] for d in datapoints]
        # trg = tri.Triangulation(x,y,triangles=triang)
        # colors = [1.0 for _ in trg.triangles]
        # plot1 = plt.tripcolor(trg,colors,antialiased=True,cmap='Blues_r',shading='flat')

        plot1.axes.set_xlabel(ax_name[0])
        plot1.axes.set_ylabel(ax_name[1])
        plot1.axes.set_xlim(ax_limits[0][0] * 1.05, ax_limits[0][1] * 1.05)
        plot1.axes.set_ylim(ax_limits[1][0] * 1.05, ax_limits[1][1] * 1.05)
        if show:
            try:
                plt.show()
            except UserWarning as e:
                if 'FigureCanvasTemplate is non-interactive' in str(e):
                    logging.warning('warning: Interactive plot not supported in current execution environment.')
        return datapoints, triang, plot1

    elif num_axes == 3:
        max_diff_y = max([abs(l - u) for l, u in zip(lb, ub)])
        datapoints = []
        datapoints_top = []
        datapoints_bottom = []
        for i, (x, l, u) in enumerate(zip(x_space.copy(), lb, ub)):
            if l != u:
                y_space = linspace(l, u, int(-(-points // (max_diff_y / abs(l - u)))))
            else:
                y_space = [l]
            datapoints_top += [[]]
            datapoints_bottom += [[]]
            for j, y in enumerate(y_space):
                constr = kwargs[CONSTRAINTS].copy()
                if ax_type[0] == 'rate':
                    constr += [[axes[0][0], '=', x]]
                elif ax_type[0] == 'yield':
                    constr += [[{**axes[0][0], **{k: -v * x for k, v in axes[0][1].items()}}, '=', 0]]
                if ax_type[1] == 'rate':
                    constr += [[axes[1][0], '=', y]]
                elif ax_type[1] == 'yield':
                    constr += [[{**axes[1][0], **{k: -v * y for k, v in axes[1][1].items()}}, '=', 0]]
                if ax_type[2] == 'rate':
                    sol_vmin = fba(model, constraints=constr, obj=axes[2][0], obj_sense='minimize')
                    sol_vmax = fba(model, constraints=constr, obj=axes[2][0], obj_sense='maximize')
                elif ax_type[2] == 'yield':
                    sol_vmin = yopt(model, constraints=constr, obj_num=axes[2][0], obj_den=axes[2][1], obj_sense='minimize')
                    sol_vmax = yopt(model, constraints=constr, obj_num=axes[2][0], obj_den=axes[2][1], obj_sense='maximize')
                datapoints_top[-1] += [len(datapoints)]
                datapoints += [array([x, y, floor_dec(sol_vmax.objective_value, 9)])]
                datapoints_bottom[-1] += [len(datapoints)]
                datapoints += [array([x, y, ceil_dec(sol_vmin.objective_value, 9)])]
            if any([isnan(datapoints[d][2]) for d in datapoints_top[-1] + datapoints_bottom[-1]]):
                logging.warning('warning: An optimization finished infeasible. Some sample points are missing.')
                datapoints_top = datapoints_top[:-1]
                datapoints_bottom = datapoints_bottom[:-1]
                x_space.remove(x)

        # Construct Denaunay triangles for plotting from all 6 perspectives
        triang = []
        # triangles top
        for i in range(len(datapoints_top) - 1):
            temp_points = datapoints_top[i] + datapoints_top[i + 1]
            pts = [array([datapoints[idx_p][0], datapoints[idx_p][1]]) for idx_p in temp_points]
            if matrix_rank(pts - pts[0]) > 1:
                triang_temp = Delaunay(pts).simplices
                triang += [[temp_points[idx] for idx in p] for p in triang_temp]
        # triangles bottom
        for i in range(len(datapoints_bottom) - 1):
            temp_points = datapoints_bottom[i] + datapoints_bottom[i + 1]
            pts = [array([datapoints[idx_p][0], datapoints[idx_p][1]]) for idx_p in temp_points]
            if matrix_rank(pts - pts[0]) > 1:
                triang_temp = Delaunay(pts).simplices
                triang += [[temp_points[idx] for idx in flip(p)] for p in triang_temp]
        # triangles front
        for i in range(len(x_space) - 1):
            temp_points = [datapoints_top[i][0], datapoints_top[i + 1][0], datapoints_bottom[i][0], datapoints_bottom[i + 1][0]]
            pts = [array([datapoints[idx_p][0], datapoints[idx_p][2]]) for idx_p in temp_points]
            if matrix_rank(pts - pts[0]) > 1:
                triang_temp = Delaunay(pts).simplices
                triang += [[temp_points[idx] for idx in flip(p)] for p in triang_temp]
        # triangles back
        for i in range(len(x_space) - 1):
            temp_points = [datapoints_top[i][-1], datapoints_top[i + 1][-1], datapoints_bottom[i][-1], datapoints_bottom[i + 1][-1]]
            pts = [array([datapoints[idx_p][0], datapoints[idx_p][2]]) for idx_p in temp_points]
            if matrix_rank(pts - pts[0]) > 1:
                triang_temp = Delaunay(pts).simplices
                triang += [[temp_points[idx] for idx in flip(p)] for p in triang_temp]
        # triangles left
        for i in range(len(datapoints_top[0]) - 1):
            temp_points = [datapoints_top[0][i], datapoints_top[0][i + 1], datapoints_bottom[0][i], datapoints_bottom[0][i + 1]]
            pts = [array([datapoints[idx_p][1], datapoints[idx_p][2]]) for idx_p in temp_points]
            if matrix_rank(pts - pts[0]) > 1:
                triang_temp = Delaunay(pts).simplices
                triang += [[temp_points[idx] for idx in p] for p in triang_temp]
        # triangles right
        for i in range(len(datapoints_top[-1]) - 1):
            temp_points = [datapoints_top[-1][i], datapoints_top[-1][i + 1], datapoints_bottom[-1][i], datapoints_bottom[-1][i + 1]]
            pts = [array([datapoints[idx_p][1], datapoints[idx_p][2]]) for idx_p in temp_points]
            if matrix_rank(pts - pts[0]) > 1:
                triang_temp = Delaunay(pts).simplices
                triang += [[temp_points[idx] for idx in p] for p in triang_temp]

        # hull = ConvexHull(datapoints)
        x = [d[0] for d in datapoints]
        y = [d[1] for d in datapoints]
        z = [d[2] for d in datapoints]
        # ax = a3.Axes3D(plt.figure())
        ax = plt.figure().add_subplot(projection='3d')
        ax.dist = 10
        ax.azim = 30
        ax.elev = 10
        ax.set_xlim(ax_limits[0])
        ax.set_ylim(ax_limits[1])
        ax.set_zlim(ax_limits[2])
        ax.set_xlabel(ax_name[0])
        ax.set_ylabel(ax_name[1])
        ax.set_zlabel(ax_name[2])
        # set higher color value for 'larger' values in all dimensions
        # use middle value of triangles instead of means and then norm
        colors = [nan for _ in triang]
        for i, t in enumerate(triang):
            p = [datapoints[i] for i in t]
            interior_point = [0.0, 0.0, 0.0]
            for d in range(3):
                v = [q[d] for q in p]
                interior_point[d] = mean([max(v), min(v)]) / max(abs(array(ax_limits[d])))
            colors[i] = sum(interior_point)  # norm(interior_point)
        lw = min([1, 6.0 / len(triang)])
        plot1 = ax.plot_trisurf(x, y, z, triangles=triang, linewidth=lw, edgecolors='black', antialiased=True,
                                alpha=0.90)  #  array=colors, cmap=plt.cm.winter
        colors = colors / max(colors)
        colors = plt.get_cmap("Spectral")(colors)
        plot1.set_fc(colors)
        if show:
            try:
                plt.show()
            except UserWarning as e:
                if 'FigureCanvasTemplate is non-interactive' in str(e):
                    logging.warning('warning: Interactive plot not supported in current execution environment.')
        return datapoints, triang, plot1


def ceil_dec(v, n):
    """Round up v to n decimals"""
    return ceil(v * (10**n)) / (10**n)


def floor_dec(v, n):
    """Round down v to n decimals"""
    return floor(v * (10**n)) / (10**n)
